import numpy as np
from joblib import (
    Parallel,
    delayed,
)
from sklearn.base import (
    MultiOutputMixin,
    RegressorMixin,
)
from sklearn.metrics import check_scoring
from sklearn.model_selection import KFold


class RidgeRegression2FoldCV(MultiOutputMixin, RegressorMixin):
    r"""Ridge regression with an efficient 2-fold cross-validation method using the SVD solver.
    Minimizes the objective function:

    .. math::

        \|y - Xw\|^2_2 + \alpha \|w\|^2_2,

    while the alpha value is determined with a 2-fold cross-validation from a list of
    alpha values. It is more efficient than doing a 2-fold cross-validation using
    :obj:`sklearn.linear_model.RidgeCV`.
    The advantage over :obj:`sklearn.linear_model.RidgeCV` using leave-one-out cross-validation
    (LOOCV) [loocv]_ needs to be analyzed more in detail. Internal benchmarks suggest that it
    is more efficient than the LOOCV in :obj:`sklearn.linear_model.RidgeCV` for feature sizes < 600
    and in general more accurate, see issue #40. However, it is constraint to a svd
    solver for the matrix inversion.
    It offers additional functionalities in comparison to :obj:`sklearn.linear_model.Ridge`:
    The regularaization parameters can be chosen relative to the largest eigenvalue of the feature matrix
    as well as regularization method. Details are explained in the `Parameters` section.

    Parameters
    ----------
    alphas : ndarray of shape (n_alphas,), default=(0.1, 1.0, 10.0)
        Array of alpha values to try.
        Regularization strength; must be a positive float. Regularization
        improves the conditioning of the problem and reduces the variance of
        the estimates. Larger values specify stronger regularization.
    alpha_type : str, default="absolute"
        "absolute" assumes that `alphas` are absolute values and does not scale them,
        "relative" assumes that `alphas` as relative values in the range [0,1] and
        scales them with respect to the largest eigenvalue.
    regularization_method : str, default="tikhonov"
        "tikhonov" uses the ``alphas`` with standard Tihhonov regularization,
        "cutoff" uses the ``alphas as a cutoff for the eigenvalues of the kernel,
        The case "cutoff" with "relative" ``alphas`` has the same effect as the `rcond`
        parameter in e.g. :obj:`numpy.linalg.lstsq`. Be aware that for every case
        we always apply a small default cutoff dependend on the numerical
        accuracy of the data type of ``X`` in the fitting function.
    random_state : int or RandomState instance, default=None
        Controls the shuffling applied to the data before applying the split.
        Pass an int for reproducible output across multiple function calls.
        See
        `random_state glossary from sklearn (external link) <https://scikit-learn.org/stable/glossary.html#term-random-state>`_
    shuffle : bool, default=True
        Whether or not to shuffle the data before splitting. If shuffle=False
        then stratify must be None.
    scoring : str, callable, default=None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.
        If None, the negative mean squared error is used.
    n_jobs : int, default=None
        The number of CPUs to use to do the computation.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See
        `n_jobs glossary from sklearn (external link) <https://scikit-learn.org/stable/glossary.html#term-n-jobs>`_
        for more details.

    Attributes
    ----------
    cv_values_ : ndarray of shape (n_samples, n_alphas) or \
        shape (n_samples, n_targets, n_alphas), optional
        Cross-validation values for each alpha (only available if \
        ``store_cv_values=True`` and ``cv=None``). After ``fit()`` has been \
        called, this attribute will contain the mean squared errors \
        (by default) or the values of the ``{loss,score}_func`` function \
        (if provided in the constructor).
    coef_ : ndarray of shape (n_features) or (n_targets, n_features)
        Weight vector(s).
    intercept_ : float or ndarray of shape (n_targets,)
        Independent term in decision function. Set to 0.0 if
        ``fit_intercept = False``.
    alpha_ : float
        Estimated regularization parameter.
    best_score_ : float
        Score of base estimator with best alpha.

    References
    ----------
    .. [loocv] Rifkin "Regularized Least Squares."
            https://www.mit.edu/~9.520/spring07/Classes/rlsslides.pdf
    """

    def __init__(
        self,
        alphas=(0.1, 1.0, 10.0),
        alpha_type="absolute",
        regularization_method="tikhonov",
        scoring=None,
        random_state=None,
        shuffle=True,
        n_jobs=None,
    ):
        self.alphas = np.asarray(alphas)
        self.alpha_type = alpha_type
        self.regularization_method = regularization_method
        self.scoring = scoring
        self.random_state = random_state
        self.shuffle = shuffle
        self.n_jobs = n_jobs

    def fit(self, X, y):
        """
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        y : ndarray of shape (n_samples, n_targets)
            Training data, where n_samples is the number of samples
            and n_targets is the number of target properties.
        """
        # check input parameters, can be moved at some point to a sklearn-like check function
        if self.regularization_method not in ["tikhonov", "cutoff"]:
            raise ValueError(
                f"regularization method {self.regularization_method} is not known."
            )
        if self.alpha_type not in ["absolute", "relative"]:
            raise ValueError(f"alpha type {self.alpha_type} is not known.")
        if self.alpha_type == "relative" and (
            np.any(self.alphas < 0) or np.any(self.alphas >= 1)
        ):
            raise ValueError(
                "relative alphas type used, but the alphas are not within the range [0,1)"
            )

        # check_scoring uses estimators scoring function if the scorer is None, this is intercepted here
        if self.scoring is None:
            scorer = check_scoring(
                self, scoring="neg_root_mean_squared_error", allow_none=False
            )
        else:
            scorer = check_scoring(self, scoring=self.scoring, allow_none=False)
        fold1_idx, fold2_idx = next(
            KFold(
                n_splits=2, shuffle=self.shuffle, random_state=self.random_state
            ).split(X)
        )
        self.coef_ = self._2fold_cv(X, y, fold1_idx, fold2_idx, scorer)
        return self

    def predict(self, X):
        """
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        """
        return X @ self.coef_.T

    def _2fold_cv(self, X, y, fold1_idx, fold2_idx, scorer):
        # two-way split of the dataset
        X_fold1 = X[fold1_idx]
        X_fold2 = X[fold2_idx]
        y_fold1 = y[fold1_idx]
        y_fold2 = y[fold2_idx]

        U_fold1, s_fold1, Vt_fold1 = np.linalg.svd(X_fold1, full_matrices=False)
        U_fold2, s_fold2, Vt_fold2 = np.linalg.svd(X_fold2, full_matrices=False)

        # scipy.linalg.pinv default rcond value
        rcond = max(X.shape) * np.spacing(X.real.dtype.type(1))

        # cutoff for singular values
        n_fold1 = sum(s_fold1 > rcond)
        n_fold2 = sum(s_fold2 > rcond)

        # computes intermediates in the least squares solution by SVD,
        # y2 ~ X_fold2 @ (V_fold1@S_fold1@U_fold1.T)@y_fold1
        Ut_fold1_y_fold1 = U_fold1.T[:n_fold1] @ y_fold1
        Ut_fold2_y_fold2 = U_fold2.T[:n_fold2] @ y_fold2
        X_fold2_V_fold1 = X_fold2 @ Vt_fold1.T[:, :n_fold1]
        X_fold1_V_fold2 = X_fold1 @ Vt_fold2.T[:, :n_fold2]

        # The scorer want an object that will make the predictions but
        # they are already computed efficiently. This
        # identity_estimator will just return them
        identity_estimator = _IdentityRegressor()

        # we use the the maximum eigenvalue of both fold, to simplify the code
        scaled_alphas = np.copy(self.alphas)
        if self.alpha_type == "relative":
            scaled_alphas *= max(np.max(s_fold1), np.max(s_fold2))

        def _2fold_loss_cutoff(alpha):
            # error approximating X2 a-fitted model and vice versa
            n_alpha_fold1 = min(n_fold1, sum(s_fold1 > alpha))
            loss_1_to_2 = scorer(
                identity_estimator,
                y_fold2,
                (X_fold2_V_fold1[:, :n_alpha_fold1] / s_fold1[:n_alpha_fold1])
                @ Ut_fold1_y_fold1[:n_alpha_fold1],
            )
            n_alpha_fold2 = min(n_fold2, sum(s_fold2 > alpha))
            loss_2_to_1 = scorer(
                identity_estimator,
                y_fold1,
                (X_fold1_V_fold2[:, :n_alpha_fold2] / s_fold2[:n_alpha_fold2])
                @ Ut_fold2_y_fold2[:n_alpha_fold2],
            )
            return (loss_1_to_2 + loss_2_to_1) / 2

        def _2fold_loss_tikhonov(alpha):
            # error approximating X2 a-fitted model and vice versa
            loss_1_to_2 = scorer(
                identity_estimator,
                y_fold2,
                (
                    X_fold2_V_fold1
                    * (s_fold1[:n_fold1] / (s_fold1[:n_fold1] ** 2 + alpha))
                )
                @ Ut_fold1_y_fold1,
            )
            loss_2_to_1 = scorer(
                identity_estimator,
                y_fold1,
                (
                    X_fold1_V_fold2
                    * (s_fold2[:n_fold2] / (s_fold2[:n_fold2] ** 2 + alpha))
                )
                @ Ut_fold2_y_fold2,
            )
            return (loss_1_to_2 + loss_2_to_1) / 2

        if self.regularization_method == "tikhonov":
            _2fold_loss_function = _2fold_loss_tikhonov
        elif self.regularization_method == "cutoff":
            _2fold_loss_function = _2fold_loss_cutoff

        self.cv_values_ = Parallel(n_jobs=self.n_jobs)(
            delayed(_2fold_loss_function)(alpha) for alpha in scaled_alphas
        )
        self.best_score_ = np.max(self.cv_values_)
        best_alpha_idx = np.argmax(self.cv_values_)
        self.alpha_ = self.alphas[best_alpha_idx]
        best_scaled_alpha = scaled_alphas[best_alpha_idx]

        U, s, Vt = np.linalg.svd(X, full_matrices=False)
        n = len(s > rcond)
        if self.regularization_method == "tikhonov":
            return (
                (Vt.T[:, :n] * s[:n] / (s[:n] ** 2 + best_scaled_alpha)) @ (U.T[:n] @ y)
            ).T
        elif self.regularization_method == "cutoff":
            n_alpha = min(n, sum(s > best_scaled_alpha))
            return ((Vt.T[:, :n_alpha] / s[:n_alpha]) @ (U.T[:n_alpha] @ y)).T


class _IdentityRegressor:
    """Fake regressor which will directly output the prediction."""

    def predict(self, y_predict):
        return y_predict



"""
The :mod:`skmatter.decomposition` module includes the two distance
measures, as defined by Principal Covariates Regression (PCovR)
"""

from ._kernel_pcovr import KernelPCovR
from ._pcovr import (
    PCovR,
    pcovr_covariance,
    pcovr_kernel,
)

__all__ = ["pcovr_covariance", "pcovr_kernel", "PCovR", "KernelPCovR"]



import numbers
import warnings

import numpy as np
from numpy.linalg import LinAlgError
from scipy import linalg
from scipy.linalg import sqrtm as MatrixSqrt
from scipy.sparse.linalg import svds
from sklearn.decomposition._base import _BasePCA
from sklearn.decomposition._pca import _infer_dimension
from sklearn.linear_model import (
    LinearRegression,
    Ridge,
    RidgeCV,
)
from sklearn.linear_model._base import LinearModel
from sklearn.utils import (
    check_array,
    check_random_state,
)
from sklearn.utils._arpack import _init_arpack_v0
from sklearn.utils.extmath import (
    randomized_svd,
    stable_cumsum,
    svd_flip,
)
from sklearn.utils.validation import (
    check_is_fitted,
    check_X_y,
)

from ..utils import (
    check_lr_fit,
    pcovr_covariance,
    pcovr_kernel,
)


class PCovR(_BasePCA, LinearModel):
    r"""

    Principal Covariates Regression, as described in [deJong1992]_
    determines a latent-space projection :math:`\mathbf{T}` which
    minimizes a combined loss in supervised and unsupervised tasks.

    This projection is determined by the eigendecomposition of a modified gram
    matrix :math:`\mathbf{\tilde{K}}`

    .. math::

      \mathbf{\tilde{K}} = \alpha \mathbf{X} \mathbf{X}^T +
            (1 - \alpha) \mathbf{\hat{Y}}\mathbf{\hat{Y}}^T

    where :math:`\alpha` is a mixing parameter and
    :math:`\mathbf{X}` and :math:`\mathbf{\hat{Y}}` are matrices of shapes
    :math:`(n_{samples}, n_{features})` and :math:`(n_{samples}, n_{properties})`,
    respectively, which contain the input and approximate targets. For
    :math:`(n_{samples} < n_{features})`, this can be more efficiently computed
    using the eigendecomposition of a modified covariance matrix
    :math:`\mathbf{\tilde{C}}`

    .. math::

      \mathbf{\tilde{C}} = \alpha \mathbf{X}^T \mathbf{X} +
            (1 - \alpha) \left(\left(\mathbf{X}^T
            \mathbf{X}\right)^{-\frac{1}{2}} \mathbf{X}^T
            \mathbf{\hat{Y}}\mathbf{\hat{Y}}^T \mathbf{X} \left(\mathbf{X}^T
            \mathbf{X}\right)^{-\frac{1}{2}}\right)

    For all PCovR methods, it is strongly suggested that :math:`\mathbf{X}` and
    :math:`\mathbf{Y}` are centered and scaled to unit variance, otherwise the
    results will change drastically near :math:`\alpha \to 0` and :math:`\alpha \to 0`.
    This can be done with the companion preprocessing classes, where

    >>> from skmatter.preprocessing import StandardFlexibleScaler as SFS
    >>>
    >>> # Set column_wise to True when the columns are relative to one another,
    >>> # False otherwise.
    >>> scaler = SFS(column_wise=True)
    >>>
    >>> scaler.fit(A) # replace with your matrix
    >>> A = scaler.transform(A)

    Parameters
    ----------
    mixing: float, default=0.5
        mixing parameter, as described in PCovR as :math:`{\alpha}`, here named
        to avoid confusion with regularization parameter `alpha`

    n_components : int, float or str, default=None
        Number of components to keep.
        if n_components is not set all components are kept::

            n_components == min(n_samples, n_features)

    svd_solver : {'auto', 'full', 'arpack', 'randomized'}, default='auto'
        If auto :
            The solver is selected by a default policy based on `X.shape` and
            `n_components`: if the input data is larger than 500x500 and the
            number of components to extract is lower than 80% of the smallest
            dimension of the data, then the more efficient 'randomized'
            method is enabled. Otherwise the exact full SVD is computed and
            optionally truncated afterwards.
        If full :
            run exact full SVD calling the standard LAPACK solver via
            `scipy.linalg.svd` and select the components by postprocessing
        If arpack :
            run SVD truncated to n_components calling ARPACK solver via
            `scipy.sparse.linalg.svds`. It requires strictly
            0 < n_components < min(X.shape)
        If randomized :
            run randomized SVD by the method of Halko et al.

    tol : float, default=1e-12
        Tolerance for singular values computed by svd_solver == 'arpack'.
        Must be of range [0.0, infinity).

    space: {'feature', 'sample', 'auto'}, default='auto'
            whether to compute the PCovR in `sample` or `feature` space
            default=`sample` when :math:`{n_{samples} < n_{features}}` and
            `feature` when :math:`{n_{features} < n_{samples}}`

    regressor: {`Ridge`, `RidgeCV`, `LinearRegression`, `precomputed`}, default=None
             regressor for computing approximated :math:`{\mathbf{\hat{Y}}}`.
             The regressor should be one `sklearn.linear_model.Ridge`,
             `sklearn.linear_model.RidgeCV`, or `sklearn.linear_model.LinearRegression`.
             If a pre-fitted regressor is provided, it is used to compute :math:`{\mathbf{\hat{Y}}}`.
             Note that any pre-fitting of the regressor will be lost if `PCovR` is
             within a composite estimator that enforces cloning, e.g.,
             `sklearn.compose.TransformedTargetRegressor` or
             `sklearn.pipeline.Pipeline` with model caching.
             In such cases, the regressor will be re-fitted on the same
             training data as the composite estimator.
             If `precomputed`, we assume that the `y` passed to the `fit` function
             is the regressed form of the targets :math:`{\mathbf{\hat{Y}}}`.
             If None, `sklearn.linear_model.Ridge('alpha':1e-6, 'fit_intercept':False, 'tol':1e-12)` is used as the regressor.

    iterated_power : int or 'auto', default='auto'
         Number of iterations for the power method computed by
         svd_solver == 'randomized'.
         Must be of range [0, infinity).

    random_state : int, RandomState instance or None, default=None
         Used when the 'arpack' or 'randomized' solvers are used. Pass an int
         for reproducible results across multiple function calls.

    Attributes
    ----------

    mixing: float, default=0.5
        mixing parameter, as described in PCovR as :math:`{\alpha}`

    tol: float, default=1e-12
        Tolerance for singular values computed by svd_solver == 'arpack'.
        Must be of range [0.0, infinity).

    space: {'feature', 'sample', 'auto'}, default='auto'
            whether to compute the PCovR in `sample` or `feature` space
            default=`sample` when :math:`{n_{samples} < n_{features}}` and
            `feature` when :math:`{n_{features} < n_{samples}}`

    n_components : int
        The estimated number of components, which equals the parameter
        n_components, or the lesser value of n_features and n_samples
        if n_components is None.

    pxt_ : ndarray of size :math:`({n_{samples}, n_{components}})`
           the projector, or weights, from the input space :math:`\mathbf{X}`
           to the latent-space projection :math:`\mathbf{T}`

    pty_ : ndarray of size :math:`({n_{components}, n_{properties}})`
          the projector, or weights, from the latent-space projection
          :math:`\mathbf{T}` to the properties :math:`\mathbf{Y}`

    pxy_ : ndarray of size :math:`({n_{samples}, n_{properties}})`
           the projector, or weights, from the input space :math:`\mathbf{X}`
           to the properties :math:`\mathbf{Y}`

    explained_variance_ : ndarray of shape (n_components,)
        The amount of variance explained by each of the selected components.

        Equal to n_components largest eigenvalues
        of the PCovR-modified covariance matrix of :math:`\mathbf{X}`.

    singular_values_ : ndarray of shape (n_components,)
        The singular values corresponding to each of the selected components.

    Examples
    --------
    >>> import numpy as np
    >>> from skmatter.decomposition import PCovR
    >>> X = np.array([[-1, 1, -3, 1], [1, -2, 1, 2], [-2, 0, -2, -2], [1, 0, 2, -1]])
    >>> Y = np.array([[ 0, -5], [-1, 1], [1, -5], [-3, 2]])
    >>> pcovr = PCovR(mixing=0.1, n_components=2)
    >>> pcovr.fit(X, Y)
    PCovR(mixing=0.1, n_components=2, space='sample')
    >>> T = pcovr.transform(X)
        [[ 3.2630561 ,  0.06663787],
         [-2.69395511, -0.41582771],
         [ 3.48683147, -0.83164387],
         [-4.05593245,  1.18083371]]
    >>> Yp = pcovr.predict(X)
        [[ 0.01374656, -5.00943466],
         [-1.02804032,  1.06737777],
         [ 0.98167556, -4.9830631 ],
         [-2.99627428,  1.98241962]]
    """

    def __init__(
        self,
        mixing=0.5,
        n_components=None,
        svd_solver="auto",
        tol=1e-12,
        space="auto",
        regressor=None,
        iterated_power="auto",
        random_state=None,
    ):
        self.mixing = mixing
        self.n_components = n_components
        self.space = space

        self.whiten = False
        self.svd_solver = svd_solver
        self.tol = tol
        self.iterated_power = iterated_power
        self.random_state = random_state

        self.regressor = regressor

    def fit(self, X, Y, W=None):
        r"""

        Fit the model with X and Y. Depending on the dimensions of X,
        calls either `_fit_feature_space` or `_fit_sample_space`

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and
            n_features is the number of features.

            It is suggested that :math:`\mathbf{X}` be centered by its column-
            means and scaled. If features are related, the matrix should be scaled
            to have unit variance, otherwise :math:`\mathbf{X}` should be
            scaled so that each feature has a variance of 1 / n_features.

        Y : ndarray, shape (n_samples, n_properties)
            Training data, where n_samples is the number of samples and
            n_properties is the number of properties

            It is suggested that :math:`\mathbf{X}` be centered by its column-
            means and scaled. If features are related, the matrix should be scaled
            to have unit variance, otherwise :math:`\mathbf{Y}` should be
            scaled so that each feature has a variance of 1 / n_features.

            If the passed regressor = `precomputed`, it is assumed that Y is the
            regressed form of the properties, :math:`{\mathbf{\hat{Y}}}`.

        W : ndarray, shape (n_features, n_properties)
            Regression weights, optional when regressor=`precomputed`. If not
            passed, it is assumed that `W = np.linalg.lstsq(X, Y, self.tol)[0]`

        """

        X, Y = check_X_y(X, Y, y_numeric=True, multi_output=True)

        # saved for inverse transformations from the latent space,
        # should be zero in the case that the features have been properly centered
        self.mean_ = np.mean(X, axis=0)

        if np.max(np.abs(self.mean_)) > self.tol:
            warnings.warn(
                "This class does not automatically center data, and your data mean is greater than the supplied tolerance."
            )

        if self.space is not None and self.space not in [
            "feature",
            "sample",
            "auto",
        ]:
            raise ValueError("Only feature and sample space are supported.")

        # Handle self.n_components==None
        if self.n_components is None:
            if self.svd_solver != "arpack":
                self.n_components = min(X.shape)
            else:
                self.n_components = min(X.shape) - 1

        if not any(
            [
                self.regressor is None,
                self.regressor == "precomputed",
                isinstance(self.regressor, LinearRegression),
                isinstance(self.regressor, Ridge),
                isinstance(self.regressor, RidgeCV),
            ]
        ):
            raise ValueError(
                "Regressor must be an instance of "
                "`LinearRegression`, `Ridge`, `RidgeCV`, or `precomputed`"
            )

        # Assign the default regressor
        if self.regressor != "precomputed":
            if self.regressor is None:
                regressor = Ridge(
                    alpha=1e-6,
                    fit_intercept=False,
                    tol=1e-12,
                )
            else:
                regressor = self.regressor

            self.regressor_ = check_lr_fit(regressor, X, y=Y)

            W = self.regressor_.coef_.T.reshape(X.shape[1], -1)
            Yhat = self.regressor_.predict(X).reshape(X.shape[0], -1)
        else:
            Yhat = Y.copy()
            if W is None:
                W = np.linalg.lstsq(X, Yhat, self.tol)[0]

        # Handle svd_solver
        self._fit_svd_solver = self.svd_solver
        if self._fit_svd_solver == "auto":
            # Small problem or self.n_components == 'mle', just call full PCA
            if max(X.shape) <= 500 or self.n_components == "mle":
                self._fit_svd_solver = "full"
            elif self.n_components >= 1 and self.n_components < 0.8 * min(X.shape):
                self._fit_svd_solver = "randomized"
            # This is also the case of self.n_components in (0,1)
            else:
                self._fit_svd_solver = "full"

        self.n_samples_, self.n_features_ = X.shape
        if self.space is None or self.space == "auto":
            if self.n_samples_ > self.n_features_:
                self.space = "feature"
            else:
                self.space = "sample"

        if self.space == "feature":
            self._fit_feature_space(X, Y.reshape(Yhat.shape), Yhat)
        else:
            self._fit_sample_space(X, Y.reshape(Yhat.shape), Yhat, W)

        self.pxy_ = self.pxt_ @ self.pty_
        if len(Y.shape) == 1:
            self.pxy_ = self.pxy_.reshape(
                X.shape[1],
            )
            self.pty_ = self.pty_.reshape(
                self.n_components,
            )

        self.components_ = self.pxt_.T  # for sklearn compatibility
        return self

    def _fit_feature_space(self, X, Y, Yhat):
        r"""
        In feature-space PCovR, the projectors are determined by:

        .. math::

            \mathbf{\tilde{C}} = \alpha \mathbf{X}^T \mathbf{X} +
            (1 - \alpha) \left(\left(\mathbf{X}^T
            \mathbf{X}\right)^{-\frac{1}{2}} \mathbf{X}^T
            \mathbf{\hat{Y}}\mathbf{\hat{Y}}^T \mathbf{X} \left(\mathbf{X}^T
            \mathbf{X}\right)^{-\frac{1}{2}}\right)

        where

        .. math::

            \mathbf{P}_{XT} = (\mathbf{X}^T \mathbf{X})^{-\frac{1}{2}}
                                \mathbf{U}_\mathbf{\tilde{C}}^T
                                \mathbf{\Lambda}_\mathbf{\tilde{C}}^{\frac{1}{2}}

        .. math::

            \mathbf{P}_{TX} = \mathbf{\Lambda}_\mathbf{\tilde{C}}^{-\frac{1}{2}}
                                \mathbf{U}_\mathbf{\tilde{C}}^T
                                (\mathbf{X}^T \mathbf{X})^{\frac{1}{2}}

        .. math::

            \mathbf{P}_{TY} = \mathbf{\Lambda}_\mathbf{\tilde{C}}^{-\frac{1}{2}}
                               \mathbf{U}_\mathbf{\tilde{C}}^T (\mathbf{X}^T
                               \mathbf{X})^{-\frac{1}{2}} \mathbf{X}^T
                               \mathbf{Y}

        """

        Ct, iCsqrt = pcovr_covariance(
            mixing=self.mixing,
            X=X,
            Y=Yhat,
            rcond=self.tol,
            return_isqrt=True,
        )
        try:
            Csqrt = np.linalg.lstsq(iCsqrt, np.eye(len(iCsqrt)), rcond=None)[0]

        # if we can avoid recomputing Csqrt, we should, but sometimes we
        # run into a singular matrix, which is what we do here
        except LinAlgError:
            Csqrt = np.real(MatrixSqrt(X.T @ X))

        if self._fit_svd_solver == "full":
            U, S, Vt = self._decompose_full(Ct)
        elif self._fit_svd_solver in ["arpack", "randomized"]:
            U, S, Vt = self._decompose_truncated(Ct)
        else:
            raise ValueError(
                "Unrecognized svd_solver='{0}'" "".format(self._fit_svd_solver)
            )

        self.singular_values_ = np.sqrt(S.copy())
        self.explained_variance_ = S / (X.shape[0] - 1)
        self.explained_variance_ratio_ = (
            self.explained_variance_ / self.explained_variance_.sum()
        )

        S_sqrt = np.diagflat([np.sqrt(s) if s > self.tol else 0.0 for s in S])
        S_sqrt_inv = np.diagflat([1.0 / np.sqrt(s) if s > self.tol else 0.0 for s in S])
        self.pxt_ = np.linalg.multi_dot([iCsqrt, Vt.T, S_sqrt])
        self.ptx_ = np.linalg.multi_dot([S_sqrt_inv, Vt, Csqrt])
        self.pty_ = np.linalg.multi_dot([S_sqrt_inv, Vt, iCsqrt, X.T, Y])

    def _fit_sample_space(self, X, Y, Yhat, W):
        r"""
        In sample-space PCovR, the projectors are determined by:

        .. math::

            \mathbf{\tilde{K}} = \alpha \mathbf{X} \mathbf{X}^T +
            (1 - \alpha) \mathbf{\hat{Y}}\mathbf{\hat{Y}}^T

        where

        .. math::

            \mathbf{P}_{XT} = \left(\alpha \mathbf{X}^T + (1 - \alpha)
                               \mathbf{W} \mathbf{\hat{Y}}^T\right)
                               \mathbf{U}_\mathbf{\tilde{K}}
                               \mathbf{\Lambda}_\mathbf{\tilde{K}}^{-\frac{1}{2}}

        .. math::

            \mathbf{P}_{TX} = \mathbf{\Lambda}_\mathbf{\tilde{K}}^{-\frac{1}{2}}
                                \mathbf{U}_\mathbf{\tilde{K}}^T \mathbf{X}

        .. math::

            \mathbf{P}_{TY} = \mathbf{\Lambda}_\mathbf{\tilde{K}}^{-\frac{1}{2}}
                               \mathbf{U}_\mathbf{\tilde{K}}^T \mathbf{Y}

        """

        Kt = pcovr_kernel(mixing=self.mixing, X=X, Y=Yhat)

        if self._fit_svd_solver == "full":
            U, S, Vt = self._decompose_full(Kt)
        elif self._fit_svd_solver in ["arpack", "randomized"]:
            U, S, Vt = self._decompose_truncated(Kt)
        else:
            raise ValueError(
                "Unrecognized svd_solver='{0}'" "".format(self._fit_svd_solver)
            )

        self.singular_values_ = np.sqrt(S.copy())
        self.explained_variance_ = S / (X.shape[0] - 1)
        self.explained_variance_ratio_ = (
            self.explained_variance_ / self.explained_variance_.sum()
        )

        P = (self.mixing * X.T) + (1.0 - self.mixing) * W @ Yhat.T
        S_sqrt_inv = np.diagflat([1.0 / np.sqrt(s) if s > self.tol else 0.0 for s in S])
        T = Vt.T @ S_sqrt_inv

        self.pxt_ = P @ T
        self.pty_ = T.T @ Y
        self.ptx_ = T.T @ X

    def _decompose_truncated(self, mat):
        if not 1 <= self.n_components <= min(self.n_samples_, self.n_features_):
            raise ValueError(
                "n_components=%r must be between 1 and "
                "min(n_samples, n_features)=%r with "
                "svd_solver='%s'"
                % (
                    self.n_components,
                    min(self.n_samples_, self.n_features_),
                    self.svd_solver,
                )
            )
        elif not isinstance(self.n_components, numbers.Integral):
            raise ValueError(
                "n_components=%r must be of type int "
                "when greater than or equal to 1, was of type=%r"
                % (self.n_components, type(self.n_components))
            )
        elif self.svd_solver == "arpack" and self.n_components == min(
            self.n_samples_, self.n_features_
        ):
            raise ValueError(
                "n_components=%r must be strictly less than "
                "min(n_samples, n_features)=%r with "
                "svd_solver='%s'"
                % (
                    self.n_components,
                    min(self.n_samples_, self.n_features_),
                    self.svd_solver,
                )
            )

        random_state = check_random_state(self.random_state)

        if self._fit_svd_solver == "arpack":
            v0 = _init_arpack_v0(min(mat.shape), random_state)
            U, S, Vt = svds(mat, k=self.n_components, tol=self.tol, v0=v0)
            # svds doesn't abide by scipy.linalg.svd/randomized_svd
            # conventions, so reverse its outputs.
            S = S[::-1]
            # flip eigenvectors' sign to enforce deterministic output
            U, Vt = svd_flip(U[:, ::-1], Vt[::-1])

        # We have already eliminated all other solvers, so this must be "randomized"
        else:
            # sign flipping is done inside
            U, S, Vt = randomized_svd(
                mat,
                n_components=self.n_components,
                n_iter=self.iterated_power,
                flip_sign=True,
                random_state=random_state,
            )

        return U, S, Vt

    def _decompose_full(self, mat):
        if self.n_components == "mle":
            if self.n_samples_ < self.n_features_:
                raise ValueError(
                    "n_components='mle' is only supported " "if n_samples >= n_features"
                )
        elif not 0 <= self.n_components <= min(self.n_samples_, self.n_features_):
            raise ValueError(
                "n_components=%r must be between 1 and "
                "min(n_samples, n_features)=%r with "
                "svd_solver='%s'"
                % (
                    self.n_components,
                    min(self.n_samples_, self.n_features_),
                    self.svd_solver,
                )
            )
        elif self.n_components >= 1:
            if not isinstance(self.n_components, numbers.Integral):
                raise ValueError(
                    "n_components=%r must be of type int "
                    "when greater than or equal to 1, "
                    "was of type=%r" % (self.n_components, type(self.n_components))
                )

        U, S, Vt = linalg.svd(mat, full_matrices=False)

        # flip eigenvectors' sign to enforce deterministic output
        U, Vt = svd_flip(U, Vt)

        # Get variance explained by singular values
        explained_variance_ = S / (self.n_samples_ - 1)
        total_var = explained_variance_.sum()
        explained_variance_ratio_ = explained_variance_ / total_var

        # Postprocess the number of components required
        if self.n_components == "mle":
            self.n_components = _infer_dimension(explained_variance_, self.n_samples_)
        elif 0 < self.n_components < 1.0:
            # number of components for which the cumulated explained
            # variance percentage is superior to the desired threshold
            # side='right' ensures that number of features selected
            # their variance is always greater than self.n_components float
            # passed. More discussion in issue: #15669
            ratio_cumsum = stable_cumsum(explained_variance_ratio_)
            self.n_components = (
                np.searchsorted(ratio_cumsum, self.n_components, side="right") + 1
            )
        self.n_components = self.n_components
        return (
            U[:, : self.n_components],
            S[: self.n_components],
            Vt[: self.n_components],
        )

    def inverse_transform(self, T):
        r"""Transform data back to its original space.

        .. math::

            \mathbf{\hat{X}} = \mathbf{T} \mathbf{P}_{TX}
                              = \mathbf{X} \mathbf{P}_{XT} \mathbf{P}_{TX}


        Parameters
        ----------
        T : ndarray, shape (n_samples, n_components)
            Projected data, where n_samples is the number of samples
            and n_components is the number of components.

        Returns
        -------
        X_original ndarray, shape (n_samples, n_features)
        """

        if np.max(np.abs(self.mean_)) > self.tol:
            warnings.warn(
                "This class does not automatically un-center data, and your data mean is greater than the supplied tolerance,"
                "so the inverse transformation will be off by the original data mean."
            )

        return T @ self.ptx_

    def predict(self, X=None, T=None):
        """Predicts the property values using regression on X or T"""

        check_is_fitted(self, ["pxy_", "pty_"])

        if X is None and T is None:
            raise ValueError("Either X or T must be supplied.")

        if X is not None:
            X = check_array(X)
            return X @ self.pxy_
        else:
            T = check_array(T)
            return T @ self.pty_

    def transform(self, X=None):
        """
        Apply dimensionality reduction to X.

        X is projected on the first principal components as determined by the
        modified PCovR distances.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            New data, where n_samples is the number of samples
            and n_features is the number of features.

        """

        check_is_fitted(self, ["pxt_", "mean_"])

        return super().transform(X)

    def score(self, X, Y, T=None):
        r"""Return the (negative) total reconstruction error for X and Y,
        defined as:

        .. math::

            \ell_{X} = \frac{\lVert \mathbf{X} - \mathbf{T}\mathbf{P}_{TX} \rVert ^ 2}{\lVert \mathbf{X}\rVert ^ 2}

        and

        .. math::

            \ell_{Y} = \frac{\lVert \mathbf{Y} - \mathbf{T}\mathbf{P}_{TY} \rVert ^ 2}{\lVert \mathbf{Y}\rVert ^ 2}

        The negative loss :math:`-\ell = -(\ell_{X} + \ell{Y})` is returned for easier use in sklearn pipelines, e.g., a grid search, where methods named 'score' are meant to be maximized.


        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The data.

        Y : ndarray of shape (n_samples, n_properties)
            The target.

        Returns
        -------
        loss : float
             Negative sum of the loss in reconstructing X from the latent-space projection T
             and the loss in predicting Y from the latent-space projection T
        """

        if T is None:
            T = self.transform(X)

        x = self.inverse_transform(T)
        y = self.predict(T=T)

        return -(
            np.linalg.norm(X - x) ** 2.0 / np.linalg.norm(X) ** 2.0
            + np.linalg.norm(Y - y) ** 2.0 / np.linalg.norm(Y) ** 2.0
        )



import unittest

import numpy as np
from sklearn.datasets import load_diabetes as get_dataset

from skmatter.sample_selection import PCovCUR

EPSILON = 1e-6


class TestPCovCUR(unittest.TestCase):
    def setUp(self):
        self.X, self.y = get_dataset(return_X_y=True)
        self.idx = [256, 304, 58, 10, 23, 278, 230, 285, 291, 357]

    def test_known(self):
        """
        This test checks that the model returns a known set of indices
        """

        selector = PCovCUR(n_to_select=10, mixing=0.5)
        selector.fit(self.X, self.y)

        self.assertTrue(np.allclose(selector.selected_idx_, self.idx))

    def test_restart(self):
        """
        This test checks that the model can be restarted with a new instance
        """

        selector = PCovCUR(n_to_select=1, mixing=0.5)
        selector.fit(self.X, self.y)

        for i in range(len(self.idx) - 2):
            selector.n_to_select += 1
            selector.fit(self.X, self.y, warm_start=True)
            self.assertEqual(selector.selected_idx_[i], self.idx[i])

            self.assertLessEqual(
                np.linalg.norm(selector.X_current_[self.idx[i]]), EPSILON
            )

            for j in range(self.X.shape[0]):
                self.assertLessEqual(
                    np.dot(selector.X_current_[self.idx[i]], selector.X_current_[j]),
                    EPSILON,
                )

    def test_non_it(self):
        """
        This test checks that the model can be run non-iteratively
        """
        self.idx = [256, 32, 138, 290, 362, 141, 359, 254, 428, 9]
        selector = PCovCUR(n_to_select=10, recompute_every=0)
        selector.fit(self.X, self.y)

        self.assertTrue(np.allclose(selector.selected_idx_, self.idx))

    def test_multiple_k(self):
        """
        This test checks that the model can be run with multiple k's
        """

        for k in np.logspace(0, np.log10(self.X.shape[0]), 4, dtype=int):
            selector = PCovCUR(n_to_select=10, k=k)
            selector.fit(self.X, self.y)


if __name__ == "__main__":
    unittest.main(verbosity=2)



import unittest
import warnings

import numpy as np
from sklearn import exceptions
from sklearn.datasets import load_diabetes as get_dataset
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import Ridge
from sklearn.kernel_ridge import KernelRidge
from sklearn.utils.validation import check_X_y

from skmatter.decomposition import PCovR


class PCovRBaseTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model = lambda mixing=0.5, regressor=Ridge(
            alpha=1e-8, fit_intercept=False, tol=1e-12
        ), **kwargs: PCovR(mixing, regressor=regressor, **kwargs)
        self.error_tol = 1e-5

        self.X, self.Y = get_dataset(return_X_y=True)
        self.X = StandardScaler().fit_transform(self.X)
        self.Y = StandardScaler().fit_transform(np.vstack(self.Y))

    def setUp(self):
        pass


class PCovRErrorTest(PCovRBaseTest):
    def test_against_pca(self):
        """
        Tests that mixing = 1.0 corresponds to PCA
        """

        pcovr = PCovR(
            mixing=1.0, n_components=3, space="sample", svd_solver="full"
        ).fit(self.X, self.Y)
        pca = PCA(n_components=3, svd_solver="full").fit(self.X)

        # tests that the SVD is equivalent
        self.assertTrue(np.allclose(pca.singular_values_, pcovr.singular_values_))
        self.assertTrue(np.allclose(pca.explained_variance_, pcovr.explained_variance_))

        T_pcovr = pcovr.transform(self.X)
        T_pca = pca.transform(self.X)

        # tests that the projections are equivalent
        self.assertLessEqual(
            np.linalg.norm(T_pcovr @ T_pcovr.T - T_pca @ T_pca.T), 1e-8
        )

    def test_simple_reconstruction(self):
        """
        This test checks that PCovR with a full eigendecomposition at mixing=1
        can fully reconstruct the input matrix.
        """

        for space in ["feature", "sample", "auto"]:
            with self.subTest(space=space):
                pcovr = self.model(
                    mixing=1.0, n_components=self.X.shape[-1], space=space
                )
                pcovr.fit(self.X, self.Y)
                Xr = pcovr.inverse_transform(pcovr.transform(self.X))
                self.assertLessEqual(
                    np.linalg.norm(self.X - Xr) ** 2.0 / np.linalg.norm(self.X) ** 2.0,
                    self.error_tol,
                )

    def test_simple_prediction(self):
        """
        This test checks that PCovR with a full eigendecomposition at mixing=0
        can fully reconstruct the input properties.
        """
        for space in ["feature", "sample", "auto"]:
            with self.subTest(space=space):
                pcovr = self.model(mixing=0.0, n_components=1, space=space)

                pcovr.regressor.fit(self.X, self.Y)
                Yhat = pcovr.regressor.predict(self.X)

                pcovr.fit(self.X, self.Y)
                Yp = pcovr.predict(self.X)
                self.assertLessEqual(
                    np.linalg.norm(Yp - Yhat) ** 2.0 / np.linalg.norm(Yhat) ** 2.0,
                    self.error_tol,
                )

    def test_lr_with_x_errors(self):
        """
        This test checks that PCovR returns a non-null property prediction
        and that the prediction error increases with `mixing`
        """
        prev_error = -1.0

        for i, mixing in enumerate(np.linspace(0, 1, 11)):
            pcovr = self.model(mixing=mixing, n_components=2, tol=1e-12)
            pcovr.fit(self.X, self.Y)

            Yp = pcovr.predict(X=self.X)
            error = np.linalg.norm(self.Y - Yp) ** 2.0 / np.linalg.norm(self.Y) ** 2.0

            with self.subTest(error=error):
                self.assertFalse(np.isnan(error))
            with self.subTest(error=error, alpha=round(mixing, 4)):
                self.assertGreaterEqual(error, prev_error - self.error_tol)

            prev_error = error

    def test_lr_with_t_errors(self):
        """
        This test checks that PCovR returns a non-null property prediction
        from the latent space projection and that the prediction error
        increases with `mixing`
        """

        prev_error = -1.0

        for i, mixing in enumerate(np.linspace(0, 1, 11)):
            pcovr = self.model(mixing=mixing, n_components=2, tol=1e-12)
            pcovr.fit(self.X, self.Y)

            T = pcovr.transform(self.X)
            Yp = pcovr.predict(T=T)
            error = np.linalg.norm(self.Y - Yp) ** 2.0 / np.linalg.norm(self.Y) ** 2.0

            with self.subTest(error=error):
                self.assertFalse(np.isnan(error))
            with self.subTest(error=error, alpha=round(mixing, 4)):
                self.assertGreaterEqual(error, prev_error - self.error_tol)

            prev_error = error

    def test_reconstruction_errors(self):
        """
        This test checks that PCovR returns a non-null reconstructed X
        and that the reconstruction error decreases with `mixing`
        """

        prev_error = 1.0

        for i, mixing in enumerate(np.linspace(0, 1, 11)):
            pcovr = self.model(mixing=mixing, n_components=2, tol=1e-12)
            pcovr.fit(self.X, self.Y)

            Xr = pcovr.inverse_transform(pcovr.transform(self.X))
            error = np.linalg.norm(self.X - Xr) ** 2.0 / np.linalg.norm(self.X) ** 2.0

            with self.subTest(error=error):
                self.assertFalse(np.isnan(error))
            with self.subTest(error=error, alpha=round(mixing, 4)):
                self.assertLessEqual(error, prev_error + self.error_tol)

            prev_error = error


class PCovRSpaceTest(PCovRBaseTest):
    def test_select_feature_space(self):
        """
        This test checks that PCovR implements the feature space version
        when :math:`n_{features} < n_{samples}``.
        """
        pcovr = self.model(n_components=2, tol=1e-12)
        pcovr.fit(self.X, self.Y)

        self.assertTrue(pcovr.space == "feature")

    def test_select_sample_space(self):
        """
        This test checks that PCovR implements the sample space version
        when :math:`n_{features} > n_{samples}``.
        """
        pcovr = self.model(n_components=2, tol=1e-12)

        n_samples = self.X.shape[1] - 1
        pcovr.fit(self.X[:n_samples], self.Y[:n_samples])

        self.assertTrue(pcovr.space == "sample")

    def test_bad_space(self):
        """
        This test checks that PCovR raises a ValueError when a non-valid
        space is designated.
        """
        with self.assertRaises(ValueError):
            pcovr = self.model(n_components=2, tol=1e-12, space="bad")
            pcovr.fit(self.X, self.Y)

    def test_override_spaceselection(self):
        """
        This test checks that PCovR implements the space provided in the
        constructor, overriding that chosen by the input dimensions.
        """
        pcovr = self.model(n_components=2, tol=1e-12, space="sample")
        pcovr.fit(self.X, self.Y)

        self.assertTrue(pcovr.space == "sample")

    def test_spaces_equivalent(self):
        """
        This test checks that the results from PCovR, regardless of the space,
        are equivalent.
        """
        for alpha in np.linspace(0.01, 0.99, 11):
            with self.subTest(alpha=alpha, type="prediction"):
                pcovr_ss = self.model(
                    n_components=2, mixing=alpha, tol=1e-12, space="sample"
                )
                pcovr_ss.fit(self.X, self.Y)

                pcovr_fs = self.model(
                    n_components=2, mixing=alpha, tol=1e-12, space="feature"
                )
                pcovr_fs.fit(self.X, self.Y)

                self.assertTrue(
                    np.allclose(
                        pcovr_ss.predict(self.X),
                        pcovr_fs.predict(self.X),
                        self.error_tol,
                    )
                )

            with self.subTest(alpha=alpha, type="reconstruction"):
                pcovr_ss = self.model(
                    n_components=2, mixing=alpha, tol=1e-12, space="sample"
                )
                pcovr_ss.fit(self.X, self.Y)

                pcovr_fs = self.model(
                    n_components=2, mixing=alpha, tol=1e-12, space="feature"
                )
                pcovr_fs.fit(self.X, self.Y)

                self.assertTrue(
                    np.allclose(
                        pcovr_ss.inverse_transform(pcovr_ss.transform(self.X)),
                        pcovr_fs.inverse_transform(pcovr_fs.transform(self.X)),
                        self.error_tol,
                    )
                )


class PCovRTestSVDSolvers(PCovRBaseTest):
    def test_svd_solvers(self):
        """
        This test checks that PCovR works with all svd_solver modes and assigns
        the right n_components
        """
        for solver in ["arpack", "full", "randomized", "auto"]:
            with self.subTest(solver=solver):
                pcovr = self.model(tol=1e-12, svd_solver=solver)
                pcovr.fit(self.X, self.Y)

                if solver == "arpack":
                    self.assertTrue(pcovr.n_components == min(self.X.shape) - 1)
                else:
                    self.assertTrue(pcovr.n_components == min(self.X.shape))

    def test_bad_solver(self):
        """
        This test checks that PCovR will not work with a solver that isn't in
        ['arpack', 'full', 'randomized', 'auto']
        """
        for space in ["feature", "sample"]:
            with self.assertRaises(ValueError) as cm:
                pcovr = self.model(svd_solver="bad", space=space)
                pcovr.fit(self.X, self.Y)

            self.assertEqual(str(cm.exception), "Unrecognized svd_solver='bad'" "")

    def test_good_n_components(self):
        """
        This test checks that PCovR will work with any allowed values of
        n_components.
        """

        # this one should pass
        pcovr = self.model(n_components=0.5, svd_solver="full")
        pcovr.fit(self.X, self.Y)

        for svd_solver in ["auto", "full"]:
            # this one should pass
            pcovr = self.model(n_components=2, svd_solver=svd_solver)
            pcovr.fit(self.X, self.Y)

            # this one should pass
            pcovr = self.model(n_components="mle", svd_solver=svd_solver)
            pcovr.fit(self.X, self.Y)

    def test_bad_n_components(self):
        """
        This test checks that PCovR will not work with any prohibited values of
        n_components.
        """

        with self.assertRaises(ValueError) as cm:
            pcovr = self.model(n_components="mle", svd_solver="full")
            pcovr.fit(self.X[:2], self.Y[:2])
        self.assertEqual(
            str(cm.exception),
            "n_components='mle' is only supported " "if n_samples >= n_features",
        )

        with self.subTest(type="negative_ncomponents"):
            with self.assertRaises(ValueError) as cm:
                pcovr = self.model(n_components=-1, svd_solver="auto")
                pcovr.fit(self.X, self.Y)

            self.assertEqual(
                str(cm.exception),
                "n_components=%r must be between 1 and "
                "min(n_samples, n_features)=%r with "
                "svd_solver='%s'"
                % (
                    pcovr.n_components,
                    min(self.X.shape),
                    pcovr.svd_solver,
                ),
            )
        with self.subTest(type="0_ncomponents"):
            with self.assertRaises(ValueError) as cm:
                pcovr = self.model(n_components=0, svd_solver="randomized")
                pcovr.fit(self.X, self.Y)

            self.assertEqual(
                str(cm.exception),
                "n_components=%r must be between 1 and "
                "min(n_samples, n_features)=%r with "
                "svd_solver='%s'"
                % (
                    pcovr.n_components,
                    min(self.X.shape),
                    pcovr.svd_solver,
                ),
            )
        with self.subTest(type="arpack_X_ncomponents"):
            with self.assertRaises(ValueError) as cm:
                pcovr = self.model(n_components=min(self.X.shape), svd_solver="arpack")
                pcovr.fit(self.X, self.Y)
            self.assertEqual(
                str(cm.exception),
                "n_components=%r must be strictly less than "
                "min(n_samples, n_features)=%r with "
                "svd_solver='%s'"
                % (
                    pcovr.n_components,
                    min(self.X.shape),
                    pcovr.svd_solver,
                ),
            )

        for svd_solver in ["auto", "full"]:
            with self.subTest(type="pi_ncomponents"):
                with self.assertRaises(ValueError) as cm:
                    pcovr = self.model(n_components=np.pi, svd_solver=svd_solver)
                    pcovr.fit(self.X, self.Y)
                self.assertEqual(
                    str(cm.exception),
                    "n_components=%r must be of type int "
                    "when greater than or equal to 1, was of type=%r"
                    % (pcovr.n_components, type(pcovr.n_components)),
                )


class PCovRInfrastructureTest(PCovRBaseTest):
    def test_nonfitted_failure(self):
        """
        This test checks that PCovR will raise a `NonFittedError` if
        `transform` is called before the pcovr is fitted
        """
        pcovr = self.model(n_components=2, tol=1e-12)
        with self.assertRaises(exceptions.NotFittedError):
            _ = pcovr.transform(self.X)

    def test_no_arg_predict(self):
        """
        This test checks that PCovR will raise a `ValueError` if
        `predict` is called without arguments
        """
        pcovr = self.model(n_components=2, tol=1e-12)
        pcovr.fit(self.X, self.Y)
        with self.assertRaises(ValueError):
            _ = pcovr.predict()

    def test_centering(self):
        """
        This test checks that PCovR raises a warning if
        given uncentered data.
        """
        pcovr = self.model(n_components=2, tol=1e-12)
        X = self.X.copy() + np.random.uniform(-1, 1, self.X.shape[1])
        with warnings.catch_warnings(record=True) as w:
            pcovr.fit(X, self.Y)
            self.assertEquals(
                str(w[0].message),
                "This class does not automatically center data, and your data mean is greater than the supplied tolerance.",
            )

    def test_T_shape(self):
        """
        This test checks that PCovR returns a latent space projection
        consistent with the shape of the input matrix
        """
        n_components = 5
        pcovr = self.model(n_components=n_components, tol=1e-12)
        pcovr.fit(self.X, self.Y)
        T = pcovr.transform(self.X)
        self.assertTrue(check_X_y(self.X, T, multi_output=True))
        self.assertTrue(T.shape[-1] == n_components)

    def test_default_ncomponents(self):
        pcovr = PCovR(mixing=0.5)
        pcovr.fit(self.X, self.Y)

        self.assertEqual(pcovr.n_components, min(self.X.shape))

    def test_Y_Shape(self):
        pcovr = self.model()
        self.Y = np.vstack(self.Y)
        pcovr.fit(self.X, self.Y)

        self.assertEqual(pcovr.pxy_.shape[0], self.X.shape[1])
        self.assertEqual(pcovr.pty_.shape[0], pcovr.n_components)

    def test_prefit_regressor(self):
        regressor = Ridge(alpha=1e-8, fit_intercept=False, tol=1e-12)
        regressor.fit(self.X, self.Y)
        pcovr = self.model(mixing=0.5, regressor=regressor)
        pcovr.fit(self.X, self.Y)

        Yhat_regressor = regressor.predict(self.X).reshape(self.X.shape[0], -1)
        W_regressor = regressor.coef_.T.reshape(self.X.shape[1], -1)

        Yhat_pcovr = pcovr.regressor_.predict(self.X).reshape(self.X.shape[0], -1)
        W_pcovr = pcovr.regressor_.coef_.T.reshape(self.X.shape[1], -1)

        self.assertTrue(np.allclose(Yhat_regressor, Yhat_pcovr))
        self.assertTrue(np.allclose(W_regressor, W_pcovr))

    def test_prefit_regression(self):
        regressor = Ridge(alpha=1e-8, fit_intercept=False, tol=1e-12)
        regressor.fit(self.X, self.Y)
        Yhat = regressor.predict(self.X)
        W = regressor.coef_.reshape(self.X.shape[1], -1)

        pcovr1 = self.model(mixing=0.5, regressor="precomputed", n_components=1)
        pcovr1.fit(self.X, Yhat, W)
        t1 = pcovr1.transform(self.X)

        pcovr2 = self.model(mixing=0.5, regressor=regressor, n_components=1)
        pcovr2.fit(self.X, self.Y)
        t2 = pcovr2.transform(self.X)

        self.assertTrue(np.linalg.norm(t1 - t2) < self.error_tol)

    def test_regressor_modifications(self):
        regressor = Ridge(alpha=1e-8)
        pcovr = self.model(mixing=0.5, regressor=regressor)

        # PCovR regressor matches the original
        self.assertTrue(regressor.get_params() == pcovr.regressor.get_params())

        # PCovR regressor updates its parameters
        # to match the original regressor
        regressor.set_params(alpha=1e-6)
        self.assertTrue(regressor.get_params() == pcovr.regressor.get_params())

        # Fitting regressor outside PCovR fits the PCovR regressor
        regressor.fit(self.X, self.Y)
        self.assertTrue(hasattr(pcovr.regressor, "coef_"))

        # PCovR regressor doesn't change after fitting
        pcovr.fit(self.X, self.Y)
        regressor.set_params(alpha=1e-4)
        self.assertTrue(hasattr(pcovr.regressor_, "coef_"))
        self.assertTrue(regressor.get_params() != pcovr.regressor_.get_params())

    def test_incompatible_regressor(self):
        regressor = KernelRidge(alpha=1e-8, kernel="linear")
        regressor.fit(self.X, self.Y)
        pcovr = self.model(mixing=0.5, regressor=regressor)

        with self.assertRaises(ValueError) as cm:
            pcovr.fit(self.X, self.Y)
        self.assertEqual(
            str(cm.exception),
            "Regressor must be an instance of "
            "`LinearRegression`, `Ridge`, `RidgeCV`, "
            "or `precomputed`",
        )

    def test_none_regressor(self):
        pcovr = PCovR(mixing=0.5, regressor=None)
        pcovr.fit(self.X, self.Y)
        self.assertTrue(pcovr.regressor is None)
        self.assertTrue(pcovr.regressor_ is not None)

    def test_incompatible_coef_shape(self):
        # self.Y is 2D with one target
        # Don't need to test X shape, since this should
        # be caught by sklearn's _validate_data
        regressor = Ridge(alpha=1e-8, fit_intercept=False, tol=1e-12)
        regressor.fit(self.X, self.Y)
        pcovr = self.model(mixing=0.5, regressor=regressor)

        # Dimension mismatch
        with self.assertRaises(ValueError) as cm:
            pcovr.fit(self.X, self.Y.squeeze())
        self.assertEqual(
            str(cm.exception),
            "The regressor coefficients have a dimension incompatible "
            "with the supplied target space. "
            "The coefficients have dimension %d and the targets "
            "have dimension %d" % (regressor.coef_.ndim, self.Y.squeeze().ndim),
        )

        # Shape mismatch (number of targets)
        with self.assertRaises(ValueError) as cm:
            pcovr.fit(self.X, np.column_stack((self.Y, self.Y)))
        self.assertEqual(
            str(cm.exception),
            "The regressor coefficients have a shape incompatible "
            "with the supplied target space. "
            "The coefficients have shape %r and the targets "
            "have shape %r"
            % (regressor.coef_.shape, np.column_stack((self.Y, self.Y)).shape),
        )


if __name__ == "__main__":
    unittest.main(verbosity=2)



"""
The :mod:`sklearn.preprocessing` module includes scaling, centering and
normalization methods.
"""

from ._data import (
    KernelNormalizer,
    SparseKernelCenterer,
    StandardFlexibleScaler,
)

__all__ = ["StandardFlexibleScaler", "KernelNormalizer", "SparseKernelCenterer"]



import unittest

from sklearn.datasets import load_diabetes as get_dataset
from sklearn.utils.validation import NotFittedError

from skmatter.feature_selection import FPS


class TestFPS(unittest.TestCase):
    def setUp(self):
        self.X, _ = get_dataset(return_X_y=True)
        self.idx = [0, 6, 1, 2, 4, 9, 3]

    def test_restart(self):
        """
        This test checks that the model can be restarted with a new number of
        features and `warm_start`
        """
        selector = FPS(n_to_select=1, initialize=self.idx[0])
        selector.fit(self.X)

        for i in range(2, len(self.idx)):
            selector.n_to_select = i
            selector.fit(self.X, warm_start=True)
            self.assertEqual(selector.selected_idx_[i - 1], self.idx[i - 1])

    def test_initialize(self):
        """
        This test checks that the model can be initialized in all applicable manners
        and throws an error otherwise
        """

        for initialize in [self.idx[0], "random"]:
            with self.subTest(initialize=initialize):
                selector = FPS(n_to_select=1, initialize=initialize)
                selector.fit(self.X)

        initialize = self.idx[:4]
        with self.subTest(initialize=initialize):
            selector = FPS(n_to_select=len(self.idx) - 1, initialize=initialize)
            selector.fit(self.X)
            for i in range(4):
                self.assertEqual(selector.selected_idx_[i], self.idx[i])

        with self.assertRaises(ValueError) as cm:
            selector = FPS(n_to_select=1, initialize="bad")
            selector.fit(self.X)
        self.assertEqual(str(cm.exception), "Invalid value of the initialize parameter")

    def test_get_distances(self):
        """
        This test checks that the haussdorf distances are returnable after fitting
        """
        selector = FPS(n_to_select=7)
        selector.fit(self.X)
        d = selector.get_select_distance()

        dist_grad = d[1:-1] - d[2:]
        self.assertTrue(all(dist_grad > 0))

        with self.assertRaises(NotFittedError):
            selector = FPS(n_to_select=7)
            _ = selector.get_select_distance()


if __name__ == "__main__":
    unittest.main(verbosity=2)



import unittest

import numpy as np
from parameterized import parameterized
from sklearn.datasets import load_iris
from sklearn.utils import (
    check_random_state,
    extmath,
)

from skmatter.linear_model import (
    OrthogonalRegression,
    RidgeRegression2FoldCV,
)


class BaseTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.features_all = load_iris().data
        cls.features_small = cls.features_all[:, [0, 1]]
        cls.features_large = cls.features_all[:, [0, 1, 0, 1]]
        cls.eps = 1e-9
        random_state = 0
        random_state = check_random_state(random_state)
        random_orthonormal_mat = extmath.randomized_range_finder(
            np.eye(cls.features_small.shape[1]),
            size=cls.features_small.shape[1],
            n_iter=10,
            random_state=random_state,
        )
        cls.features_rotated_small = cls.features_small @ random_orthonormal_mat

    def test_orthogonal_regression_small_to_rotated_small(self):
        # tests if OrthogonalRegression can predict rotated small features using small features with use_orthogonal_projector False
        err = np.linalg.norm(
            self.features_rotated_small
            - OrthogonalRegression(use_orthogonal_projector=False)
            .fit(self.features_small, self.features_rotated_small)
            .predict(self.features_small)
        )
        self.assertTrue(
            abs(err) < self.eps, f"error {err} surpasses threshold for zero {self.eps}"
        )

    def test_orthogonal_regression_large_to_small(self):
        # tests if prediction is padded to larger feature size
        n_features = (
            OrthogonalRegression(use_orthogonal_projector=False)
            .fit(self.features_large, self.features_small)
            .predict(self.features_large)
            .shape[1]
        )
        self.assertTrue(
            n_features == self.features_large.shape[1],
            f"n_features {n_features} does not match larger feature size {self.features_large.shape[1]}",
        )

    def test_orthogonal_regression_use_orthogonal_projector_small_to_rotated_small(
        self,
    ):
        # tests if OrthogonalRegression can predict rotated small features using small features with use_orthogonal_projector True
        err = np.linalg.norm(
            self.features_rotated_small
            - OrthogonalRegression(use_orthogonal_projector=True)
            .fit(self.features_small, self.features_rotated_small)
            .predict(self.features_small)
        )
        self.assertTrue(
            abs(err) < self.eps, f"error {err} surpasses threshold for zero {self.eps}"
        )

    def test_orthogonal_regression_use_orthogonal_projector_small_to_large(self):
        # tests if prediction is projected to prediction feature space
        n_features = (
            OrthogonalRegression(use_orthogonal_projector=True)
            .fit(self.features_small, self.features_large)
            .predict(self.features_small)
            .shape[1]
        )
        self.assertTrue(
            n_features == self.features_large.shape[1],
            f"n_features {n_features} does not match projection feature size {self.features_large.shape[1]}",
        )

    def test_orthogonal_regression_use_orthogonal_projector_large_to_small(self):
        # tests if prediction is projected to prediction feature space
        n_features = (
            OrthogonalRegression(use_orthogonal_projector=True)
            .fit(self.features_large, self.features_small)
            .predict(self.features_large)
            .shape[1]
        )
        self.assertTrue(
            n_features == self.features_small.shape[1],
            f"n_features {n_features} does not match projection feature size {self.features_small.shape[1]}",
        )


class RidgeTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.features_all = load_iris().data
        cls.features_small = cls.features_all[:, [0, 1]]
        cls.features_large = cls.features_all[:, [0, 1, 0, 1]]
        cls.eps = 5e-8
        np.random.RandomState(0).seed(0x5F3759DF)
        cls.alphas = [1e-9, 1e-3, 1e-1, 0.5]
        cls.ridge_regressions = []

    def test_ridge_regression_2fold_regularization_method_raise_error(self):
        # tests if wrong regularization_method in RidgeRegression2FoldCV raises error
        with self.assertRaises(ValueError):
            RidgeRegression2FoldCV(
                regularization_method="dummy",
            ).fit(self.features_small, self.features_small)

    def test_ridge_regression_2fold_alpha_type_raise_error(self):
        # tests if wrong alpha type in RidgeRegression2FoldCV raises error
        with self.assertRaises(ValueError):
            RidgeRegression2FoldCV(
                alpha_type="dummy",
            ).fit(self.features_small, self.features_small)

    def test_ridge_regression_2fold_relative_alpha_type_raise_error(self):
        # tests if an error is raised if alpha not in [0,1)
        with self.assertRaises(ValueError):
            RidgeRegression2FoldCV(alphas=[1], alpha_type="relative").fit(
                self.features_small, self.features_small
            )

        with self.assertRaises(ValueError):
            RidgeRegression2FoldCV(alphas=[-0.1], alpha_type="relative").fit(
                self.features_small, self.features_small
            )

    ridge_parameters = [
        ["absolute_tikhonov", "absolute", "tikhonov"],
        ["absolute_cutoff", "absolute", "cutoff"],
        ["relative_tikhonov", "relative", "tikhonov"],
        ["relative_cutoff", "relative", "cutoff"],
    ]

    @parameterized.expand(ridge_parameters)
    def test_ridge_regression_2fold_cv_small_to_small(
        self, name, alpha_type, regularization_method
    ):
        # tests if RidgeRegression2FoldCV can predict small features using small features with use_orthogonal_projector False
        err = np.linalg.norm(
            self.features_small
            - RidgeRegression2FoldCV(
                alphas=self.alphas,
                alpha_type=alpha_type,
                regularization_method=regularization_method,
            )
            .fit(self.features_small, self.features_small)
            .predict(self.features_small)
        )
        self.assertTrue(
            abs(err) < self.eps, f"error {err} surpasses threshold for zero {self.eps}"
        )

    @parameterized.expand(ridge_parameters)
    def test_ridge_regression_2fold_cv_small_to_large(
        # tests if RidgeRegression2FoldCV can predict large features using small features with use_orthogonal_projector False
        self,
        name,
        alpha_type,
        regularization_method,
    ):
        err = np.linalg.norm(
            self.features_large
            - RidgeRegression2FoldCV(
                alphas=self.alphas,
                alpha_type=alpha_type,
                regularization_method=regularization_method,
            )
            .fit(self.features_small, self.features_large)
            .predict(self.features_small)
        )
        self.assertTrue(
            abs(err) < self.eps,
            f"error {err} surpasses threshold for zero {self.eps}",
        )

    @parameterized.expand(ridge_parameters)
    def test_ridge_regression_2fold_regularization(
        self, name, alpha_type, regularization_method
    ):
        # tests if the regularization in the CV split of
        # RidgeRegression2FoldCV does effect the results

        # regularization parameters are chosen to match the singular values o
        # the features, thus each regularization parameter affects the minimized
        # weight matrix and thus the error
        _, singular_values, _ = np.linalg.svd(self.features_all)
        if alpha_type == "absolute":
            alphas = singular_values[1:][::-1]
        if alpha_type == "relative":
            alphas = singular_values[1:][::-1] / singular_values[0]

        # tests if RidgeRegression2FoldCV does do regularization correct
        ridge = RidgeRegression2FoldCV(
            alphas=alphas,
            alpha_type=alpha_type,
            regularization_method=regularization_method,
            scoring="neg_root_mean_squared_error",
        ).fit(self.features_all, self.features_all)
        twofold_rmse = -np.array(ridge.cv_values_)

        # since the data can be perfectly reconstructed,
        # larger regularization parameters (alphas) should result in
        # larger errors
        error_grad = twofold_rmse[1:] - twofold_rmse[:-1]
        self.assertTrue(
            np.all(error_grad > self.eps),
            "error does not strictly increase with larger regularization\n"
            f"\ttwofold RMSE: {twofold_rmse}\n"
            f"\tregularization parameters: {ridge.alphas}",
        )


if __name__ == "__main__":
    unittest.main()



"""
The :mod:`skmatter.utils` module includes functions which are
used by multiple packages
"""

from ._orthogonalizers import (
    X_orthogonalizer,
    Y_feature_orthogonalizer,
    Y_sample_orthogonalizer,
)
from ._pcovr_utils import (
    check_krr_fit,
    check_lr_fit,
    pcovr_covariance,
    pcovr_kernel,
)
from ._progress_bar import (
    get_progress_bar,
    no_progress_bar,
)

__all__ = [
    "get_progress_bar",
    "no_progress_bar",
    "pcovr_covariance",
    "pcovr_kernel",
    "check_krr_fit",
    "check_lr_fit",
    "X_orthogonalizer",
    "Y_sample_orthogonalizer",
    "Y_feature_orthogonalizer",
]



from ._base import (
    load_csd_1000r,
    load_degenerate_CH4_manifold,
    load_nice_dataset,
    load_who_dataset,
)

__all__ = [
    "load_degenerate_CH4_manifold",
    "load_csd_1000r",
    "load_nice_dataset",
    "load_who_dataset",
]



import numpy as np
from scipy.linalg import orthogonal_procrustes
from sklearn.base import (
    MultiOutputMixin,
    RegressorMixin,
)
from sklearn.linear_model import LinearRegression


class OrthogonalRegression(MultiOutputMixin, RegressorMixin):
    r"""Orthogonal regression by solving the Procrustes problem

    Linear regression with the additional constraint that the weight matrix
    must be an orthogonal matrix/projection. It minimizes the Procrustes
    problem:

    .. math::

        \min_\Omega ||y - X\Omega\||_F \quad\mathrm{subject\ to}\quad \Omega^T\Omega=I

    Parameters
    ----------
    use_orthogonal_projector : bool, default=True
        Controls if orthogonal projectors are used to predict y fitting on X.
        If this parameter is set to False X and y are padded with zeros to the larger
        number of features of X and y. The projection method is similar
        to the procedure in the computation GFRD in the first version of
        Ref. [Goscinski2021]_. The method has been adapted obtain a full weight matrix.

        The projection can introduce nonanalytic behavior with respect to
        changes in dimensions of X for cases where X n_features > y n_targets.
        See ``examples/OrthogonalRegressionNonAnalytic_no-doc.ipynb``

    linear_estimator : object implementing fit/predict, default=None
        The linear estimator is used when `use_orthogonal_projector`
        is set to True, to compute the projection matrix

    Attributes
    ----------
    max_components_ : int
        The source X and target y are padded with zeros to match in feature/target
        dimension, when `use_orthogonal_projector` is set to False. This attribute
        is set to the maximum of the feature and target dimension.

    coef_ : ndarray of shape (n_features,) or (n_targets, n_features) or (max_components, max_components)
        Weight matrix. The shape (max_components, max_components) is used if
        `use_orthogonal_projector` is set to False.

    """

    def __init__(self, use_orthogonal_projector=True, linear_estimator=None):
        self.use_orthogonal_projector = use_orthogonal_projector
        self.linear_estimator = linear_estimator

    def fit(self, X, y):
        """
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        y : ndarray of shape (n_samples, n_targets)
            Training data, where n_samples is the number of samples
            and n_targets is the number of target properties.
        """
        if self.use_orthogonal_projector:
            # check estimator
            linear_estimator = (
                LinearRegression()
                if self.linear_estimator is None
                else self.linear_estimator
            )
            # compute orthogonal projectors
            linear_estimator.fit(X, y)
            U, _, Vt = np.linalg.svd(linear_estimator.coef_.T, full_matrices=False)
            # project X and y to same dimension
            X = X @ U
            y = y @ Vt.T
            # compute weights by solving the Procrustes problem
            self.coef_ = (U @ orthogonal_procrustes(X, y)[0] @ Vt).T
        else:
            self.max_components_ = max(X.shape[1], y.shape[1])
            X = np.pad(X, [(0, 0), (0, self.max_components_ - X.shape[1])])
            y = np.pad(y, [(0, 0), (0, self.max_components_ - y.shape[1])])
            self.coef_ = orthogonal_procrustes(X, y)[0].T

        return self

    def predict(self, X):
        """
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        """
        if not (self.use_orthogonal_projector):
            X = np.pad(X, [(0, 0), (0, self.max_components_ - X.shape[1])])
        return X @ self.coef_.T



from os.path import (
    dirname,
    join,
)

import numpy as np
from sklearn.utils import (
    Bunch,
    check_pandas_support,
)


def load_nice_dataset():
    """Load and returns NICE dataset.
    Returns
    -------
    nice_data : sklearn.utils.Bunch
      Dictionary-like object, with the following attributes:
      data : `sklearn.utils.Bunch` --
      contains the keys ``X`` and ``y``.
      Structural NICE features and energies, respectively.
      DESCR: `str` --
        The full description of the dataset.
    """

    module_path = dirname(__file__)
    target_filename = join(module_path, "data", "nice_dataset.npz")
    raw_data = np.load(target_filename)
    data = Bunch(
        X=raw_data["structural_features"],
        y=raw_data["energies"],
    )
    with open(join(module_path, "descr", "nice_dataset.rst")) as rst_file:
        fdescr = rst_file.read()
    return Bunch(data=data, DESCR=fdescr)


def load_degenerate_CH4_manifold():
    """Load and return the degenerate manifold dataset.

    Returns
    -------
    degenerate_CH4_manifold_data : sklearn.utils.Bunch
        Dictionary-like object, with the following attributes:

        data : `sklearn.utils.Bunch` --
        contains the keys ``SOAP_power_spectrum`` and ``SOAP_bispectrum``.
        Two representations of the carbon environments of the
        degenerate manifold dataset.

        DESCR: `str` --
        The full description of the dataset.
    """
    module_path = dirname(__file__)
    target_filename = join(module_path, "data", "degenerate_CH4_manifold.npz")
    raw_data = np.load(target_filename)
    data = Bunch(
        SOAP_power_spectrum=raw_data["SOAP_power_spectrum"],
        SOAP_bispectrum=raw_data["SOAP_bispectrum"],
    )
    with open(join(module_path, "descr", "degenerate_CH4_manifold.rst")) as rst_file:
        fdescr = rst_file.read()

    return Bunch(data=data, DESCR=fdescr)


def load_csd_1000r(return_X_y=False):
    """Load and return the minimal CSD dataset.

    Returns
    -------
    csd1000r : sklearn.utils.Bunch
        Dictionary-like object, with the following attributes:

        data : `sklearn.utils.Bunch` --
        contains the keys ``X`` and ``Y``, corresponding to the
        FPS-reduced SOAP vectors and local NMR chemical shielding, respectively,
        for 100 selected environments of the CSD-1000r dataset.

        DESCR: `str` --
        The full description of the dataset.
    """
    module_path = dirname(__file__)
    target_filename = join(module_path, "data", "csd-1000r.npz")
    raw_data = np.load(target_filename)
    if not return_X_y:
        data = Bunch(
            X=raw_data["X"],
            y=raw_data["Y"],
        )
        with open(join(module_path, "descr", "csd-1000r.rst")) as rst_file:
            fdescr = rst_file.read()

        return Bunch(data=data, DESCR=fdescr)
    else:
        return raw_data["X"], raw_data["Y"]


def load_who_dataset():
    """Load and returns WHO dataset.
    Returns
    -------
    who_dataset : sklearn.utils.Bunch
      Dictionary-like object, with the following attributes:
          data : `pandas.core.frame.DataFrame` -- the WHO dataset
                  as a Pandas dataframe.
          DESCR: `str` -- The full description of the dataset.
    """

    module_path = dirname(__file__)
    target_filename = join(module_path, "data", "who_dataset.csv")
    pd = check_pandas_support("load_who_dataset")
    raw_data = pd.read_csv(target_filename)
    with open(join(module_path, "descr", "who_dataset.rst")) as rst_file:
        fdescr = rst_file.read()
    return Bunch(data=raw_data, DESCR=fdescr)



import unittest

import numpy as np
from sklearn.exceptions import NotFittedError

from skmatter.sample_selection import (
    FPS,
    VoronoiFPS,
)
from test_sample_simple_fps import TestFPS


class TestVoronoiFPS(TestFPS):
    def setUp(self):
        super().setUp()

    def test_restart(self):
        """
        This test checks that the model can be restarted with a new number of
        features and `warm_start`
        """

        selector = VoronoiFPS(n_to_select=1, initialize=self.idx[0])
        selector.fit(self.X)

        for i in range(2, len(self.idx)):
            selector.n_to_select = i
            selector.fit(self.X, warm_start=True)
            self.assertEqual(selector.selected_idx_[i - 1], self.idx[i - 1])

    def test_initialize(self):
        """
        This test checks that the model can be initialized in all applicable manners
        and throws an error otherwise
        """

        for initialize in [self.idx[0], "random"]:
            with self.subTest(initialize=initialize):
                selector = VoronoiFPS(n_to_select=1, initialize=initialize)
                selector.fit(self.X)

        with self.assertRaises(ValueError) as cm:
            selector = VoronoiFPS(n_to_select=1, initialize="bad")
            selector.fit(self.X)
        self.assertEquals(
            str(cm.exception), "Invalid value of the initialize parameter"
        )

    def test_switching_point(self):
        """
        This test check work of the switching point calculator into the
        _init_greedy_search function
        """
        selector = VoronoiFPS(n_to_select=1)
        selector.fit(self.X)
        self.assertTrue(1 > selector.full_fraction)

        selector = VoronoiFPS(n_to_select=1, full_fraction=0.5)
        selector.fit(self.X)
        self.assertEqual(selector.full_fraction, 0.5)

        with self.subTest(name="bad_ntrial"):
            with self.assertRaises(ValueError) as cm:
                selector = VoronoiFPS(n_to_select=1, n_trial_calculation=0)
                selector.fit(self.X)
            self.assertEqual(
                str(cm.exception),
                "Number of trial calculation should be more or equal to 1",
            )

        with self.subTest(name="float_ntrial"):
            with self.assertRaises(TypeError) as cm:
                selector = VoronoiFPS(n_to_select=1, n_trial_calculation=0.3)
                selector.fit(self.X)
            self.assertEqual(
                str(cm.exception), "Number of trial calculation should be integer"
            )

        with self.subTest(name="large_ff"):
            with self.assertRaises(ValueError) as cm:
                selector = VoronoiFPS(n_to_select=1, full_fraction=1.1)
                selector.fit(self.X)
            self.assertEqual(
                str(cm.exception),
                "Switching point should be real and more than 0 and less than 1. "
                f"Received {selector.full_fraction}",
            )

        with self.subTest(name="string_ff"):
            with self.assertRaises(ValueError) as cm:
                selector = VoronoiFPS(n_to_select=1, full_fraction="STRING")
                selector.fit(self.X)
            self.assertEqual(
                str(cm.exception),
                "Switching point should be real and more than 0 and less than 1. "
                f"Received {selector.full_fraction}",
            )

    def test_get_distances(self):
        """
        This test checks that the haussdorf distances are returnable after fitting
        """
        selector = VoronoiFPS(n_to_select=1)
        selector.fit(self.X)
        _ = selector.get_select_distance()

        with self.assertRaises(NotFittedError):
            selector = VoronoiFPS(n_to_select=1)
            _ = selector.get_select_distance()

    def test_comparison(self):
        """
        This test checks that the voronoi FPS strictly computes less distances
        than its normal FPS counterpart.
        """
        vselector = VoronoiFPS(n_to_select=self.X.shape[0] - 1)
        vselector.fit(self.X)

        selector = FPS(n_to_select=self.X.shape[0] - 1)
        selector.fit(self.X)

        self.assertTrue(np.allclose(vselector.selected_idx_, selector.selected_idx_))

    def test_nothing_updated_points(self):
        """
        This test checks that in the case where we have no points to update,
        the code still works fine
        """
        X = np.array([[1, 1], [4, 4], [10, 10], [100, 100]])
        selector = VoronoiFPS(n_to_select=3, initialize=0)
        try:
            selector.fit(X)
            f = 1
        except Exception:
            f = 0
        self.assertEqual(f, 1)

        self.assertEqual(
            len(np.where(selector.vlocation_of_idx == (selector.n_selected_ - 2))[0]), 1
        )

    def test_calculate_dSL(self):
        selector = VoronoiFPS(n_to_select=3)
        selector.fit(self.X)

        active_points = np.where(
            selector.dSL_[selector.vlocation_of_idx] < selector.haussdorf_
        )[0]

        ap = selector._get_active(self.X, selector.selected_idx_[-1])

        self.assertTrue(
            np.allclose(
                active_points,
                ap,
            )
        )

        selector = VoronoiFPS(n_to_select=1)

        ap = selector._get_active(self.X, 0)

        self.assertTrue(
            np.allclose(
                np.arange(self.X.shape[0]),
                ap,
            )
        )

    def test_score(self):
        """This test check that function score return haussdorf distance"""
        selector = VoronoiFPS(n_to_select=3, initialize=0)
        selector.fit(self.X)

        self.assertTrue(
            np.allclose(
                selector.haussdorf_,
                selector.score(self.X, selector.selected_idx_[-1]),
            )
        )


if __name__ == "__main__":
    unittest.main(verbosity=2)



import unittest

import numpy as np
from sklearn.preprocessing import StandardScaler

from skmatter.datasets import load_csd_1000r
from skmatter.utils import (
    X_orthogonalizer,
    Y_feature_orthogonalizer,
    Y_sample_orthogonalizer,
)

EPSILON = 1e-8


class TestXOrth(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.random_state = np.random.RandomState(0)

    def setUp(self):
        self.n_samples = 100
        self.n_features = 100

    def test_null_column(self):
        # checks that the column passed to the orthogonalizer
        # is empty post-orthogonalization

        n_uncorrelated = self.n_features // 2

        X_random = self.random_state.uniform(
            -1, 1, size=(self.n_samples, self.n_features)
        )
        X_correlated = np.zeros((self.n_samples, self.n_features))
        X_correlated[:, :n_uncorrelated] = self.random_state.uniform(
            -1, 1, size=(self.n_samples, n_uncorrelated)
        )

        for i in range(n_uncorrelated, self.n_features):
            X_correlated[:, i] = X_correlated[
                :, i - n_uncorrelated
            ] * self.random_state.uniform(-1, 1)

        feat_idx = np.arange(min(self.n_samples, self.n_features, n_uncorrelated))
        self.random_state.shuffle(feat_idx)

        for idx in feat_idx:
            with self.subTest(type="random X"):
                X_random = X_orthogonalizer(X_random, c=idx)
                self.assertLessEqual(np.linalg.norm(X_random[:, idx]), EPSILON)

            with self.subTest(type="correlated X"):
                X_correlated = X_orthogonalizer(X_correlated, c=idx)
                self.assertLessEqual(np.linalg.norm(X_correlated[:, idx]), EPSILON)
                self.assertLessEqual(
                    np.linalg.norm(X_correlated[:, idx + n_uncorrelated]), EPSILON
                )

    def test_null_row(self):
        # checks that the row passed to the orthogonalizer
        # is empty post-orthogonalization

        n_uncorrelated = self.n_samples // 2

        X_random = self.random_state.uniform(
            -1, 1, size=(self.n_samples, self.n_features)
        )
        X_random2 = self.random_state.uniform(
            -1, 1, size=(self.n_samples, self.n_features)
        )
        X_correlated = np.zeros((self.n_samples, self.n_features))
        X_correlated[:n_uncorrelated] = self.random_state.uniform(
            -1, 1, size=(n_uncorrelated, self.n_features)
        )

        for i in range(n_uncorrelated, self.n_samples):
            X_correlated[i] = X_correlated[
                i - n_uncorrelated
            ] * self.random_state.uniform(-1, 1)

        feat_idx = np.arange(min(self.n_samples, self.n_features, n_uncorrelated))
        self.random_state.shuffle(feat_idx)

        for idx in feat_idx:
            with self.subTest(type="random X"):
                X_random = X_orthogonalizer(X_random.T, c=idx).T
                self.assertLessEqual(np.linalg.norm(X_random[idx]), EPSILON)

            with self.subTest(type="random X with column"):
                X_random2 = X_orthogonalizer(X_random2.T, x2=X_random2[idx].T).T
                self.assertLessEqual(np.linalg.norm(X_random2[idx]), EPSILON)

            with self.subTest(type="correlated X"):
                X_correlated = X_orthogonalizer(X_correlated.T, c=idx).T
                self.assertLessEqual(np.linalg.norm(X_correlated[idx]), EPSILON)
                self.assertLessEqual(
                    np.linalg.norm(X_correlated[idx + n_uncorrelated]), EPSILON
                )

    def test_multiple_orthogonalizations(self):
        # checks that the matrix is empty when orthogonalized simultaneously
        # by all uncorrelated columns

        n_uncorrelated = self.n_samples // 2

        X_correlated = np.zeros((self.n_samples, self.n_features))
        X_correlated[:, :n_uncorrelated] = self.random_state.uniform(
            -1, 1, size=(self.n_samples, n_uncorrelated)
        )

        for i in range(n_uncorrelated, self.n_features):
            X_correlated[:, i] = X_correlated[
                :, i - n_uncorrelated
            ] * self.random_state.uniform(-1, 1)

        X_correlated = X_orthogonalizer(
            X_correlated, x2=X_correlated[:, :n_uncorrelated]
        )

        self.assertLessEqual(np.linalg.norm(X_correlated), EPSILON)

    def test_multicolumn(self):
        # checks that an error is raised when x2 is the wrong shape for x1
        with self.assertRaises(ValueError) as cm:
            X_orthogonalizer(
                self.random_state.uniform(
                    -3, 3, size=(self.n_samples, self.n_features)
                ),
                x2=self.random_state.uniform(
                    -3, 3, size=(self.n_samples + 4, self.n_features)
                ),
            )
        self.assertEqual(
            str(cm.exception),
            "You can only orthogonalize a matrix using a vector with the same number of rows."
            f"Matrix X has {self.n_samples} rows, whereas the orthogonalizing matrix has {self.n_samples+4} rows.",
        )

    def test_warning(self):
        # checks that a warning is raised when trying to orthogonalize by
        # an empty vector
        with self.assertWarns(Warning, msg="Column vector contains only zeros."):
            X_orthogonalizer(np.zeros((self.n_samples, self.n_features)), 0)

    def test_copy(self):
        # checks that the X_orthogonalizer works in-place when copy=True

        X_random = self.random_state.uniform(
            -1, 1, size=(self.n_samples, self.n_features)
        )

        idx = self.random_state.choice(X_random.shape[-1])

        new_X = X_orthogonalizer(X_random, idx, tol=EPSILON, copy=True)
        X_orthogonalizer(X_random, idx, tol=EPSILON, copy=False)
        self.assertTrue(np.allclose(X_random, new_X))


class TestYOrths(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.random_state = np.random.RandomState(0)

    def setUp(self):
        self.X, self.y = load_csd_1000r(return_X_y=True)
        self.X = StandardScaler().fit_transform(self.X)
        self.y = StandardScaler().fit_transform(self.y)

    def test_pass_feature(self):
        # checks that the Y_feature_orthogonalizer removes all targets
        # predictable by the given set of features

        Xc = self.X[:, self.random_state.choice(self.X.shape[-1], 3)]
        yhat = Xc @ np.linalg.pinv(Xc.T @ Xc, rcond=EPSILON) @ Xc.T @ self.y

        new_y = Y_feature_orthogonalizer(self.y, Xc, tol=EPSILON)
        self.assertTrue(np.allclose(self.y - new_y, yhat))

    def test_copy_feature(self):
        # checks the Y_feature_orthogonalizer operates in-place when copy=False

        Xc = self.X[:, self.random_state.choice(self.X.shape[-1], 3)]
        new_y = Y_feature_orthogonalizer(self.y, Xc, tol=EPSILON, copy=False)
        self.assertTrue(np.allclose(self.y, new_y))

    def test_pass_sample(self):
        # checks that the Y_samples_orthogonalizer removes all targets
        # predictable by the given set of samples

        r = self.random_state.choice(self.X.shape[0], 3)
        Xr = self.X[r]
        yr = self.y[r]

        yhat = self.X @ np.linalg.pinv(Xr.T @ Xr, rcond=EPSILON) @ Xr.T @ yr

        new_y = Y_sample_orthogonalizer(self.y, self.X, yr, Xr, tol=EPSILON)
        self.assertTrue(np.allclose(self.y - new_y, yhat))

    def test_copy_sample(self):
        # checks the Y_sample_orthogonalizer operates in-place when copy=False

        r = self.random_state.choice(self.X.shape[0], 3)
        Xr = self.X[r]
        yr = self.y[r]

        new_y = Y_sample_orthogonalizer(self.y, self.X, yr, Xr, tol=EPSILON, copy=False)
        self.assertTrue(np.allclose(self.y, new_y))


if __name__ == "__main__":
    unittest.main(verbosity=2)



import unittest

import numpy as np
from sklearn import exceptions
from sklearn.datasets import load_diabetes as get_dataset
from sklearn.kernel_ridge import KernelRidge
from sklearn.linear_model import (
    Ridge,
    RidgeCV,
)
from sklearn.utils.validation import check_X_y

from skmatter.decomposition import (
    KernelPCovR,
    PCovR,
)
from skmatter.preprocessing import StandardFlexibleScaler as SFS


class KernelPCovRBaseTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.random_state = np.random.RandomState(0)

        self.error_tol = 1e-6

        self.X, self.Y = get_dataset(return_X_y=True)

        # for the sake of expedience, only use a subset of the dataset
        idx = self.random_state.choice(len(self.X), 1000)
        self.X = self.X[idx]
        self.Y = self.Y[idx]

        # artificial second property
        self.Y = np.array(
            [self.Y, self.X @ self.random_state.randint(-2, 2, (self.X.shape[-1],))]
        ).T
        self.Y = self.Y.reshape(self.X.shape[0], -1)

        self.X = SFS().fit_transform(self.X)
        self.Y = SFS(column_wise=True).fit_transform(self.Y)

        self.model = lambda mixing=0.5, regressor=KernelRidge(
            alpha=1e-8
        ), n_components=4, **kwargs: KernelPCovR(
            mixing,
            regressor=regressor,
            n_components=n_components,
            svd_solver=kwargs.pop("svd_solver", "full"),
            **kwargs
        )

    def setUp(self):
        pass


class KernelPCovRErrorTest(KernelPCovRBaseTest):
    def test_lr_with_x_errors(self):
        """
        This test checks that KernelPCovR returns a non-null property prediction
        and that the prediction error increases with `mixing`
        """
        prev_error = -1.0

        for i, mixing in enumerate(np.linspace(0, 1, 6)):
            kpcovr = KernelPCovR(mixing=mixing, n_components=2, tol=1e-12)
            kpcovr.fit(self.X, self.Y)

            error = (
                np.linalg.norm(self.Y - kpcovr.predict(self.X)) ** 2.0
                / np.linalg.norm(self.Y) ** 2.0
            )

            with self.subTest(error=error):
                self.assertFalse(np.isnan(error))
            with self.subTest(error=error, alpha=round(mixing, 4)):
                self.assertGreaterEqual(error, prev_error - self.error_tol)

            prev_error = error

    def test_reconstruction_errors(self):
        """
        This test checks that KernelPCovR returns a non-null reconstructed X
        and that the reconstruction error decreases with `mixing`
        """

        prev_error = 10.0
        prev_x_error = 10.0

        for i, mixing in enumerate(np.linspace(0, 1, 6)):
            kpcovr = KernelPCovR(
                mixing=mixing, n_components=2, fit_inverse_transform=True, tol=1e-12
            )
            kpcovr.fit(self.X, self.Y)

            t = kpcovr.transform(self.X)
            K = kpcovr._get_kernel(self.X)
            x = kpcovr.inverse_transform(t)

            error = np.linalg.norm(K - t @ t.T) ** 2.0 / np.linalg.norm(K) ** 2.0
            x_error = np.linalg.norm(self.X - x) ** 2.0 / np.linalg.norm(self.X) ** 2.0

            with self.subTest(error=error):
                self.assertFalse(np.isnan(error))
            with self.subTest(error=error, alpha=round(mixing, 4)):
                self.assertLessEqual(error, prev_error + self.error_tol)

            with self.subTest(error=x_error):
                self.assertFalse(np.isnan(x_error))
            with self.subTest(error=x_error, alpha=round(mixing, 4)):
                self.assertLessEqual(x_error, prev_x_error + self.error_tol)

            prev_error = error
            prev_x_error = x_error

    def test_kpcovr_error(self):
        for i, mixing in enumerate(np.linspace(0, 1, 6)):
            kpcovr = self.model(
                mixing=mixing,
                regressor=KernelRidge(kernel="rbf", gamma=1.0),
                kernel="rbf",
                gamma=1.0,
                center=False,
            )

            kpcovr.fit(self.X, self.Y)
            K = kpcovr._get_kernel(self.X)

            y = kpcovr.predict(self.X)
            Lkrr = np.linalg.norm(self.Y - y) ** 2 / np.linalg.norm(self.Y) ** 2

            t = kpcovr.transform(self.X)

            w = t @ np.linalg.pinv(t.T @ t, rcond=kpcovr.tol) @ t.T
            Lkpca = np.trace(K - K @ w) / np.trace(K)

            # this is only true for in-sample data
            self.assertTrue(
                np.isclose(
                    kpcovr.score(self.X, self.Y), -sum([Lkpca, Lkrr]), self.error_tol
                )
            )


class KernelPCovRInfrastructureTest(KernelPCovRBaseTest):
    def test_nonfitted_failure(self):
        """
        This test checks that KernelPCovR will raise a `NonFittedError` if
        `transform` is called before the model is fitted
        """
        kpcovr = KernelPCovR(mixing=0.5, n_components=2, tol=1e-12)
        with self.assertRaises(exceptions.NotFittedError):
            _ = kpcovr.transform(self.X)

    def test_no_arg_predict(self):
        """
        This test checks that KernelPCovR will raise a `ValueError` if
        `predict` is called without arguments
        """
        kpcovr = KernelPCovR(mixing=0.5, n_components=2, tol=1e-12)
        kpcovr.fit(self.X, self.Y)
        with self.assertRaises(ValueError):
            _ = kpcovr.predict()

    def test_T_shape(self):
        """
        This test checks that KernelPCovR returns a latent space projection
        consistent with the shape of the input matrix
        """
        n_components = 5
        kpcovr = KernelPCovR(mixing=0.5, n_components=n_components, tol=1e-12)
        kpcovr.fit(self.X, self.Y)
        T = kpcovr.transform(self.X)
        self.assertTrue(check_X_y(self.X, T, multi_output=True))
        self.assertTrue(T.shape[-1] == n_components)

    def test_no_centerer(self):
        """
        tests that when center=False, no centerer exists
        """
        kpcovr = self.model(center=False)
        kpcovr.fit(self.X, self.Y)

        with self.assertRaises(AttributeError):
            _ = getattr(kpcovr, "centerer_")

    def test_centerer(self):
        """
        tests that all functionalities that rely on the centerer work properly
        """

        kpcovr = self.model(center=True)
        kpcovr.fit(self.X, self.Y)

        self.assertTrue(hasattr(kpcovr, "centerer_"))
        _ = kpcovr.predict(self.X)
        _ = kpcovr.transform(self.X)
        _ = kpcovr.score(self.X, self.Y)

    def test_prefit_regressor(self):
        regressor = KernelRidge(alpha=1e-8, kernel="rbf", gamma=0.1)
        regressor.fit(self.X, self.Y)
        kpcovr = self.model(mixing=0.5, regressor=regressor, kernel="rbf", gamma=0.1)
        kpcovr.fit(self.X, self.Y)

        Yhat_regressor = regressor.predict(self.X).reshape(self.X.shape[0], -1)
        W_regressor = regressor.dual_coef_.reshape(self.X.shape[0], -1)

        Yhat_kpcovr = kpcovr.regressor_.predict(self.X).reshape(self.X.shape[0], -1)
        W_kpcovr = kpcovr.regressor_.dual_coef_.reshape(self.X.shape[0], -1)

        self.assertTrue(np.allclose(Yhat_regressor, Yhat_kpcovr))
        self.assertTrue(np.allclose(W_regressor, W_kpcovr))

    def test_regressor_modifications(self):
        regressor = KernelRidge(alpha=1e-8, kernel="rbf", gamma=0.1)
        kpcovr = self.model(mixing=0.5, regressor=regressor, kernel="rbf", gamma=0.1)

        # KPCovR regressor matches the original
        self.assertTrue(regressor.get_params() == kpcovr.regressor.get_params())

        # KPCovR regressor updates its parameters
        # to match the original regressor
        regressor.set_params(gamma=0.2)
        self.assertTrue(regressor.get_params() == kpcovr.regressor.get_params())

        # Fitting regressor outside KPCovR fits the KPCovR regressor
        regressor.fit(self.X, self.Y)
        self.assertTrue(hasattr(kpcovr.regressor, "dual_coef_"))

        # Raise error during KPCovR fit since regressor and KPCovR
        # kernel parameters now inconsistent
        with self.assertRaises(ValueError) as cm:
            kpcovr.fit(self.X, self.Y)
        self.assertTrue(
            str(cm.exception),
            "Kernel parameter mismatch: the regressor has kernel parameters "
            "{kernel: linear, gamma: 0.2, degree: 3, coef0: 1, kernel_params: None}"
            " and KernelPCovR was initialized with kernel parameters "
            "{kernel: linear, gamma: 0.1, degree: 3, coef0: 1, kernel_params: None}",
        )

    def test_incompatible_regressor(self):
        regressor = Ridge(alpha=1e-8)
        regressor.fit(self.X, self.Y)
        kpcovr = self.model(mixing=0.5, regressor=regressor)

        with self.assertRaises(ValueError) as cm:
            kpcovr.fit(self.X, self.Y)
        self.assertTrue(
            str(cm.exception),
            "Regressor must be an instance of `KernelRidge`",
        )

    def test_none_regressor(self):
        kpcovr = KernelPCovR(mixing=0.5, regressor=None)
        kpcovr.fit(self.X, self.Y)
        self.assertTrue(kpcovr.regressor is None)
        self.assertTrue(kpcovr.regressor_ is not None)

    def test_incompatible_coef_shape(self):
        # self.Y is 2D with two targets
        # Don't need to test X shape, since this should
        # be caught by sklearn's _validate_data
        regressor = KernelRidge(alpha=1e-8, kernel="linear")
        regressor.fit(self.X, self.Y[:, 0][:, np.newaxis])
        kpcovr = self.model(mixing=0.5, regressor=regressor)

        # Dimension mismatch
        with self.assertRaises(ValueError) as cm:
            kpcovr.fit(self.X, self.Y[:, 0])
        self.assertTrue(
            str(cm.exception),
            "The regressor coefficients have a dimension incompatible "
            "with the supplied target space. "
            "The coefficients have dimension %d and the targets "
            "have dimension %d" % (regressor.dual_coef_.ndim, self.Y[:, 0].ndim),
        )

        # Shape mismatch (number of targets)
        with self.assertRaises(ValueError) as cm:
            kpcovr.fit(self.X, self.Y)
        self.assertTrue(
            str(cm.exception),
            "The regressor coefficients have a shape incompatible "
            "with the supplied target space. "
            "The coefficients have shape %r and the targets "
            "have shape %r" % (regressor.dual_coef_.shape, self.Y.shape),
        )

    def test_precomputed_regression(self):
        regressor = KernelRidge(alpha=1e-8, kernel="rbf", gamma=0.1)
        regressor.fit(self.X, self.Y)
        Yhat = regressor.predict(self.X)
        W = regressor.dual_coef_.reshape(self.X.shape[0], -1)

        kpcovr1 = self.model(
            mixing=0.5, regressor="precomputed", kernel="rbf", gamma=0.1, n_components=1
        )
        kpcovr1.fit(self.X, Yhat, W)
        t1 = kpcovr1.transform(self.X)

        kpcovr2 = self.model(
            mixing=0.5, regressor=regressor, kernel="rbf", gamma=0.1, n_components=1
        )
        kpcovr2.fit(self.X, self.Y)
        t2 = kpcovr2.transform(self.X)

        self.assertTrue(np.linalg.norm(t1 - t2) < self.error_tol)


class KernelTests(KernelPCovRBaseTest):
    def test_kernel_types(self):
        """
        This test checks that KernelPCovR can handle all kernels passable to
        sklearn kernel classes, including callable kernels
        """

        def _linear_kernel(X, Y):
            return X @ Y.T

        kernel_params = {
            "poly": {"degree": 2},
            "rbf": {"gamma": 3.0},
            "sigmoid": {"gamma": 3.0, "coef0": 0.5},
        }
        for kernel in ["linear", "poly", "rbf", "sigmoid", "cosine", _linear_kernel]:
            with self.subTest(kernel=kernel):
                kpcovr = KernelPCovR(
                    mixing=0.5,
                    n_components=2,
                    regressor=KernelRidge(
                        kernel=kernel, **kernel_params.get(kernel, {})
                    ),
                    kernel=kernel,
                    **kernel_params.get(kernel, {})
                )
                kpcovr.fit(self.X, self.Y)

    def test_linear_matches_pcovr(self):
        """
        This test checks that KernelPCovR returns the same results as PCovR when
        using a linear kernel
        """

        ridge = RidgeCV(fit_intercept=False, alphas=np.logspace(-8, 2))
        ridge.fit(self.X, self.Y)

        # common instantiation parameters for the two models
        hypers = dict(
            mixing=0.5,
            n_components=1,
        )

        # computing projection and predicton loss with linear KernelPCovR
        # and use the alpha from RidgeCV for level regression comparisons
        kpcovr = KernelPCovR(
            regressor=KernelRidge(alpha=ridge.alpha_, kernel="linear"),
            kernel="linear",
            fit_inverse_transform=True,
            **hypers
        )
        kpcovr.fit(self.X, self.Y)
        ly = (
            np.linalg.norm(self.Y - kpcovr.predict(self.X)) ** 2.0
            / np.linalg.norm(self.Y) ** 2.0
        )

        # computing projection and predicton loss with PCovR
        ref_pcovr = PCovR(**hypers, regressor=ridge, space="sample")
        ref_pcovr.fit(self.X, self.Y)
        ly_ref = (
            np.linalg.norm(self.Y - ref_pcovr.predict(self.X)) ** 2.0
            / np.linalg.norm(self.Y) ** 2.0
        )

        t_ref = ref_pcovr.transform(self.X)
        t = kpcovr.transform(self.X)

        K = kpcovr._get_kernel(self.X)

        k_ref = t_ref @ t_ref.T
        k = t @ t.T

        lk_ref = np.linalg.norm(K - k_ref) ** 2.0 / np.linalg.norm(K) ** 2.0
        lk = np.linalg.norm(K - k) ** 2.0 / np.linalg.norm(K) ** 2.0

        rounding = 3
        self.assertEqual(
            round(ly, rounding),
            round(ly_ref, rounding),
        )

        self.assertEqual(
            round(lk, rounding),
            round(lk_ref, rounding),
        )


class KernelPCovRTestSVDSolvers(KernelPCovRBaseTest):
    def test_svd_solvers(self):
        """
        This test checks that PCovR works with all svd_solver modes and assigns
        the right n_components
        """
        for solver in ["arpack", "full", "randomized", "auto"]:
            with self.subTest(solver=solver):
                kpcovr = self.model(tol=1e-12, n_components=None, svd_solver=solver)
                kpcovr.fit(self.X, self.Y)

                if solver == "arpack":
                    self.assertTrue(kpcovr.n_components == self.X.shape[0] - 1)
                else:
                    self.assertTrue(kpcovr.n_components == self.X.shape[0])

    def test_bad_solver(self):
        """
        This test checks that PCovR will not work with a solver that isn't in
        ['arpack', 'full', 'randomized', 'auto']
        """
        with self.assertRaises(ValueError) as cm:
            kpcovr = self.model(svd_solver="bad")
            kpcovr.fit(self.X, self.Y)

        self.assertTrue(str(cm.exception), "Unrecognized svd_solver='bad'" "")

    def test_good_n_components(self):
        """
        This test checks that PCovR will work with any allowed values of
        n_components.
        """

        # this one should pass
        kpcovr = self.model(n_components=0.5, svd_solver="full")
        kpcovr.fit(self.X, self.Y)

        for svd_solver in ["auto", "full"]:
            # this one should pass
            kpcovr = self.model(n_components=2, svd_solver=svd_solver)
            kpcovr.fit(self.X, self.Y)

            # this one should pass
            kpcovr = self.model(n_components="mle", svd_solver=svd_solver)
            kpcovr.fit(self.X, self.Y)

    def test_bad_n_components(self):
        """
        This test checks that PCovR will not work with any prohibited values of
        n_components.
        """

        with self.subTest(type="negative_ncomponents"):
            with self.assertRaises(ValueError) as cm:
                kpcovr = self.model(n_components=-1, svd_solver="auto")
                kpcovr.fit(self.X, self.Y)

            self.assertTrue(
                str(cm.exception),
                "self.n_components=%r must be between 0 and "
                "min(n_samples, n_features)=%r with "
                "svd_solver='%s'"
                % (
                    kpcovr.n_components,
                    self.X.shape[0],
                    kpcovr.svd_solver,
                ),
            )
        with self.subTest(type="0_ncomponents"):
            with self.assertRaises(ValueError) as cm:
                kpcovr = self.model(n_components=0, svd_solver="randomized")
                kpcovr.fit(self.X, self.Y)

            self.assertTrue(
                str(cm.exception),
                "self.n_components=%r must be between 1 and "
                "min(n_samples, n_features)=%r with "
                "svd_solver='%s'"
                % (
                    kpcovr.n_components,
                    self.X.shape[0],
                    kpcovr.svd_solver,
                ),
            )
        with self.subTest(type="arpack_X_ncomponents"):
            with self.assertRaises(ValueError) as cm:
                kpcovr = self.model(n_components=self.X.shape[0], svd_solver="arpack")
                kpcovr.fit(self.X, self.Y)
            self.assertTrue(
                str(cm.exception),
                "self.n_components=%r must be strictly less than "
                "min(n_samples, n_features)=%r with "
                "svd_solver='%s'"
                % (
                    kpcovr.n_components,
                    self.X.shape[0],
                    kpcovr.svd_solver,
                ),
            )

        for svd_solver in ["auto", "full"]:
            with self.subTest(type="pi_ncomponents"):
                with self.assertRaises(ValueError) as cm:
                    kpcovr = self.model(n_components=np.pi, svd_solver=svd_solver)
                    kpcovr.fit(self.X, self.Y)
                self.assertTrue(
                    str(cm.exception),
                    "self.n_components=%r must be of type int "
                    "when greater than or equal to 1, was of type=%r"
                    % (kpcovr.n_components, type(kpcovr.n_components)),
                )


if __name__ == "__main__":
    unittest.main(verbosity=2)



import unittest

import numpy as np
import sklearn

from skmatter.preprocessing import KernelNormalizer


class KernelTests(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.random_state = np.random.RandomState(0)

    def test_sample_weights(self):
        """Checks that sample weights of one are equal to the unweighted case and that nonuniform weights are different from the unweighted case"""
        K = self.random_state.uniform(0, 100, size=(3, 3))
        equal_wts = np.ones(len(K))
        nonequal_wts = self.random_state.uniform(0, 100, size=(len(K),))
        model = KernelNormalizer()
        weighted_model = KernelNormalizer()
        K_unweighted = model.fit_transform(K)
        K_equal_weighted = weighted_model.fit_transform(K, sample_weight=equal_wts)
        self.assertTrue((np.isclose(K_unweighted, K_equal_weighted, atol=1e-12)).all())
        K_nonequal_weighted = weighted_model.fit_transform(
            K, sample_weight=nonequal_wts
        )
        self.assertFalse(
            (np.isclose(K_unweighted, K_nonequal_weighted, atol=1e-12)).all()
        )

    def test_invalid_sample_weights(self):
        """Checks that weights must be 1D array with the same length as the number of samples"""
        K = self.random_state.uniform(0, 100, size=(3, 3))
        wts_len = np.ones(len(K) + 1)
        wts_dim = np.ones((len(K), 2))
        model = KernelNormalizer()
        with self.assertRaises(ValueError):
            model.fit_transform(K, sample_weight=wts_len)
        with self.assertRaises(ValueError):
            model.fit_transform(K, sample_weight=wts_dim)

    def test_NoInputs(self):
        """Checks that fit cannot be called with zero inputs."""
        model = KernelNormalizer()
        with self.assertRaises(ValueError):
            model.fit()

    def test_ValueError(self):
        """Checks that a non-square matrix cannot be normalized."""
        K = self.random_state.uniform(0, 100, size=(3, 4))
        model = KernelNormalizer()
        with self.assertRaises(ValueError):
            model.fit(K)

    def test_reference_ValueError(self):
        """Checks that it is impossible to normalize
        a matrix with a non-coincident size with the reference."""
        K = self.random_state.uniform(0, 100, size=(3, 3))
        K_2 = self.random_state.uniform(0, 100, size=(2, 2))
        model = KernelNormalizer()
        model = model.fit(K)
        with self.assertRaises(ValueError):
            model.transform(K_2)

    def test_NotFittedError_transform(self):
        """Checks that an error is returned when
        trying to use the transform function
        before the fit function"""
        K = self.random_state.uniform(0, 100, size=(3, 3))
        model = KernelNormalizer()
        with self.assertRaises(sklearn.exceptions.NotFittedError):
            model.transform(K)

    def test_fit_transform(self):
        """Checks that the kernel is correctly normalized.
        Compare with the value calculated
        directly from the equation.
        """
        K = self.random_state.uniform(0, 100, size=(3, 3))
        model = KernelNormalizer()
        Ktr = model.fit_transform(K)
        Kc = K - K.mean(axis=0) - K.mean(axis=1)[:, np.newaxis] + K.mean()
        Kc /= np.trace(Kc) / Kc.shape[0]

        self.assertTrue((np.isclose(Ktr, Kc, atol=1e-12)).all())

    def test_center_only(self):
        """Checks that the kernel is correctly centered,
        but not normalized.
        Compare with the value calculated
        directly from the equation.
        """
        K = self.random_state.uniform(0, 100, size=(3, 3))
        model = KernelNormalizer(with_center=True, with_trace=False)
        Ktr = model.fit_transform(K)
        Kc = K - K.mean(axis=0) - K.mean(axis=1)[:, np.newaxis] + K.mean()

        self.assertTrue((np.isclose(Ktr, Kc, atol=1e-12)).all())

    def test_trace_only(self):
        """Checks that the kernel is correctly normalized,
        but not centered.
        Compare with the value calculated
        directly from the equation.
        """
        K = self.random_state.uniform(0, 100, size=(3, 3))
        model = KernelNormalizer(with_center=False, with_trace=True)
        Ktr = model.fit_transform(K)
        Kc = K.copy()
        Kc /= np.trace(Kc) / Kc.shape[0]

        self.assertTrue((np.isclose(Ktr, Kc, atol=1e-12)).all())

    def test_no_preprocessing(self):
        """Checks that the kernel is unchanged
        if no preprocessing is specified.
        """
        K = self.random_state.uniform(0, 100, size=(3, 3))
        model = KernelNormalizer(with_center=False, with_trace=False)
        Ktr = model.fit_transform(K)
        Kc = K.copy()
        self.assertTrue((np.isclose(Ktr, Kc, atol=1e-12)).all())


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from sklearn import exceptions

from skmatter.datasets import load_csd_1000r as load
from skmatter.feature_selection import CUR


class TestCUR(unittest.TestCase):
    def setUp(self):
        self.X, _ = load(return_X_y=True)

    def test_bad_transform(self):
        selector = CUR(n_to_select=2)
        with self.assertRaises(exceptions.NotFittedError):
            _ = selector.transform(self.X)

    def test_restart(self):
        """
        This test checks that the model can be restarted with a new instance
        """

        ref_selector = CUR(n_to_select=self.X.shape[-1] - 3).fit(X=self.X)
        ref_idx = ref_selector.selected_idx_

        selector = CUR(n_to_select=1)
        selector.fit(self.X)

        for i in range(self.X.shape[-1] - 3):
            selector.n_to_select += 1
            selector.fit(self.X, warm_start=True)
            self.assertEqual(selector.selected_idx_[i], ref_idx[i])

    def test_non_it(self):
        """
        This test checks that the model can be run non-iteratively
        """
        C = self.X.T @ self.X
        _, UC = np.linalg.eigh(C)
        ref_idx = np.argsort(-(UC[:, -1] ** 2.0))[:-1]

        selector = CUR(n_to_select=self.X.shape[-1] - 1, recompute_every=0)
        selector.fit(self.X)

        self.assertTrue(np.allclose(selector.selected_idx_, ref_idx))


if __name__ == "__main__":
    unittest.main(verbosity=2)



import sklearn.model_selection
from sklearn.utils import indexable
from sklearn.utils.validation import _num_samples


def train_test_split(*arrays, **options):
    """This is an extended version of the sklearn train test split supporting
    overlapping train and test sets.
    See `sklearn.model_selection.train_test_split (external link)
    <https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html>`_ .

    Parameters
    ----------
    *arrays : sequence of indexables with same length / shape[0]
        Allowed inputs are lists, numpy arrays, scipy-sparse
        matrices or pandas dataframes.
    test_size : float or int, default=None
        If float, should be between 0.0 and 1.0 and represent the proportion
        of the dataset to include in the test split. If int, represents the
        absolute number of test samples. If None, the value is set to the
        complement of the train size. If ``train_size`` is also None, it will
        be set to 0.25.
    train_size : float or int, default=None
        If float, should be between 0.0 and 1.0 and represent the
        proportion of the dataset to include in the train split. If
        int, represents the absolute number of train samples. If None,
        the value is automatically set to the complement of the test size.
    random_state : int or RandomState instance, default=None
        Controls the shuffling applied to the data before applying the split.
        Pass an int for reproducible output across multiple function calls.
        See
        `random state glossary from sklearn (external link) <https://scikit-learn.org/stable/glossary.html#term-random-state>`_
    shuffle : bool, default=True
        Whether or not to shuffle the data before splitting. If shuffle=False
        then stratify must be None.
    stratify : array-like, default=None
        If not None, data is split in a stratified fashion, using this as
        the class labels.
    train_test_overlap : bool, default=False
        If True, and train and test set are both not None, the train and test
        set may overlap.

    Returns
    -------
    splitting : list, length=2 * len(arrays)
        List containing train-test split of inputs.
    """
    train_test_overlap = options.pop("train_test_overlap", False)
    test_size = options.get("test_size", None)
    train_size = options.get("train_size", None)

    if train_test_overlap and train_size is not None and test_size is not None:
        # checks from sklearn
        arrays = indexable(*arrays)
        n_samples = _num_samples(arrays[0])

        if test_size == 1.0 or test_size == n_samples:
            test_sets = arrays
        else:
            options["train_size"] = None
            test_sets = sklearn.model_selection.train_test_split(*arrays, **options)[
                1::2
            ]
            options["train_size"] = train_size

        if train_size == 1.0 or train_size == n_samples:
            train_sets = arrays
        else:
            options["test_size"] = None
            train_sets = sklearn.model_selection.train_test_split(*arrays, **options)[
                ::2
            ]
            options["test_size"] = test_size

        train_test_sets = []
        for i in range(len(train_sets)):
            train_test_sets += [train_sets[i], test_sets[i]]
        return train_test_sets
    else:
        return sklearn.model_selection.train_test_split(*arrays, **options)



"""
Sequential sample selection
"""

import warnings

import numpy as np
from scipy.interpolate import (
    LinearNDInterpolator,
    interp1d,
)
from scipy.interpolate.interpnd import _ndim_coords_from_arrays
from scipy.spatial import ConvexHull
from sklearn.utils.validation import (
    check_array,
    check_is_fitted,
    check_X_y,
)

from .._selection import (
    _CUR,
    _FPS,
    _PCovCUR,
    _PCovFPS,
)


def _linear_interpolator(points, values):
    """
    Returns linear interpolater for unstructured D-D data. Tessellate the input point set to N-D simplices, and interpolate linearly on each simplex. See `LinearNDInterpolator` for more details.

    points : 2-D ndarray of floats with shape (n, D), or length D tuple of 1-D ndarrays with shape (n,).
        Data point coordinates.
    values : ndarray of float or complex, shape (n,)
        Data values.


    Reference:
    ---------
    The code is an adapted excerpt from
    https://github.com/scipy/scipy/blob/dde50595862a4f9cede24b5d1c86935c30f1f88a/scipy/interpolate/_ndgriddata.py#L119-L273
    """

    points = _ndim_coords_from_arrays(points)

    if points.ndim < 2:
        ndim = points.ndim
    else:
        ndim = points.shape[-1]

    if ndim == 1:
        points = points.ravel()
        # Sort points/values together, necessary as input for interp1d
        idx = np.argsort(points)
        points = points[idx]
        values = values[idx]
        return interp1d(
            points, values, kind="linear", axis=0, bounds_error=False, fill_value=np.nan
        )
    else:
        return LinearNDInterpolator(points, values, fill_value=np.nan, rescale=False)


class FPS(_FPS):
    """
    Transformer that performs Greedy Sample Selection using Farthest Point Sampling.

    Parameters
    ----------

    initialize: int, list of int, or 'random', default=0
        Index of the first selection(s). If 'random', picks a random
        value when fit starts. Stored in :py:attr:`self.initialize`.

    n_to_select : int or float, default=None
        The number of selections to make. If `None`, half of the samples are
        selected. If integer, the parameter is the absolute number of selections
        to make. If float between 0 and 1, it is the fraction of the total dataset to
        select. Stored in :py:attr:`self.n_to_select`.

    score_threshold : float, default=None
        Threshold for the score. If `None` selection will continue until the
        n_to_select is chosen. Otherwise will stop when the score falls below the threshold.
        Stored in :py:attr:`self.score_threshold`.

    score_threshold_type : str, default="absolute"
        How to interpret the ``score_threshold``. When "absolute", the score used by
        the selector is compared to the threshold directly. When "relative", at each iteration,
        the score used by the selector is compared proportionally to the score of the first
        selection, i.e. the selector quits when ``current_score / first_score < threshold``.
        Stored in :py:attr:`self.score_threshold_type`.

    progress_bar: bool, default=False
              option to use `tqdm <https://tqdm.github.io/>`_
              progress bar to monitor selections. Stored in :py:attr:`self.report_progress`.

    full : bool, default=False
        In the case that all non-redundant selections are exhausted, choose
        randomly from the remaining samples. Stored in :py:attr:`self.full`.

    random_state: int or RandomState instance, default=0

    Attributes
    ----------

    n_selected_ : int
                  Counter tracking the number of selections that have been made
    X_selected_ : ndarray,
                  Matrix containing the selected samples, for use in fitting
    y_selected_ : ndarray,
                  In sample selection, the matrix containing the selected targets, for use in fitting

    """

    def __init__(
        self,
        initialize=0,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        super().__init__(
            selection_type="sample",
            initialize=initialize,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            score_threshold_type=score_threshold_type,
            progress_bar=progress_bar,
            full=full,
            random_state=random_state,
        )


class PCovFPS(_PCovFPS):
    """Transformer that performs Greedy Sample Selection using PCovR-weighted
    Farthest Point Sampling.

    Parameters
    ----------

    mixing: float, default=0.5
            The PCovR mixing parameter, as described in PCovR as
            :math:`{\\alpha}`

    initialize: int or 'random', default=0
        Index of the first selection. If 'random', picks a random
        value when fit starts.

    n_to_select : int or float, default=None
        The number of selections to make. If `None`, half of the samples are
        selected. If integer, the parameter is the absolute number of selections
        to make. If float between 0 and 1, it is the fraction of the total dataset to
        select. Stored in :py:attr:`self.n_to_select`.

    score_threshold : float, default=None
        Threshold for the score. If `None` selection will continue until the
        n_to_select is chosen. Otherwise will stop when the score falls below the threshold.
        Stored in :py:attr:`self.score_threshold`.

    score_threshold_type : str, default="absolute"
        How to interpret the ``score_threshold``. When "absolute", the score used by
        the selector is compared to the threshold directly. When "relative", at each iteration,
        the score used by the selector is compared proportionally to the score of the first
        selection, i.e. the selector quits when ``current_score / first_score < threshold``.
        Stored in :py:attr:`self.score_threshold_type`.

    progress_bar: bool, default=False
              option to use `tqdm <https://tqdm.github.io/>`_
              progress bar to monitor selections. Stored in :py:attr:`self.report_progress`.

    full : bool, default=False
        In the case that all non-redundant selections are exhausted, choose
        randomly from the remaining samples. Stored in :py:attr:`self.full`.

    random_state: int or RandomState instance, default=0

    Attributes
    ----------

    n_selected_ : int
                  Counter tracking the number of selections that have been made
    X_selected_ : ndarray,
                  Matrix containing the selected samples, for use in fitting
    y_selected_ : ndarray,
                  In sample selection, the matrix containing the selected targets, for use in fitting


    """

    def __init__(
        self,
        mixing=0.5,
        initialize=0,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        super().__init__(
            selection_type="sample",
            mixing=mixing,
            initialize=initialize,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            score_threshold_type=score_threshold_type,
            progress_bar=progress_bar,
            full=full,
            random_state=random_state,
        )


class CUR(_CUR):
    """Transformer that performs Greedy Sample Selection by choosing samples
     which maximize the magnitude of the left singular vectors, consistent with
     classic CUR matrix decomposition.

    Parameters
     ----------
     recompute_every : int
                       number of steps after which to recompute the pi score
                       defaults to 1, if 0 no re-computation is done

     k : int
         number of eigenvectors to compute the importance score with, defaults to 1

     tolerance: float
          threshold below which scores will be considered 0, defaults to 1E-12

     n_to_select : int or float, default=None
         The number of selections to make. If `None`, half of the samples are
         selected. If integer, the parameter is the absolute number of selections
         to make. If float between 0 and 1, it is the fraction of the total dataset to
         select. Stored in :py:attr:`self.n_to_select`.

     score_threshold : float, default=None
         Threshold for the score. If `None` selection will continue until the
         n_to_select is chosen. Otherwise will stop when the score falls below the threshold.
         Stored in :py:attr:`self.score_threshold`.

     score_threshold_type : str, default="absolute"
         How to interpret the ``score_threshold``. When "absolute", the score used by
         the selector is compared to the threshold directly. When "relative", at each iteration,
         the score used by the selector is compared proportionally to the score of the first
         selection, i.e. the selector quits when ``current_score / first_score < threshold``.
         Stored in :py:attr:`self.score_threshold_type`.

     progress_bar: bool, default=False
               option to use `tqdm <https://tqdm.github.io/>`_
               progress bar to monitor selections. Stored in :py:attr:`self.report_progress`.

     full : bool, default=False
         In the case that all non-redundant selections are exhausted, choose
         randomly from the remaining samples. Stored in :py:attr:`self.full`.

     random_state: int or RandomState instance, default=0

     Attributes
     ----------

     X_current_ : ndarray (n_samples, n_features)
                   The original matrix orthogonalized by previous selections

     n_selected_ : int
                   Counter tracking the number of selections that have been made
     X_selected_ : ndarray,
                   Matrix containing the selected samples, for use in fitting
     y_selected_ : ndarray,
                   In sample selection, the matrix containing the selected targets, for use in fitting

    """

    def __init__(
        self,
        recompute_every=1,
        k=1,
        tolerance=1e-12,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        super().__init__(
            selection_type="sample",
            recompute_every=recompute_every,
            k=k,
            tolerance=tolerance,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            score_threshold_type=score_threshold_type,
            progress_bar=progress_bar,
            full=full,
            random_state=random_state,
        )


class PCovCUR(_PCovCUR):
    r"""Transformer that performs Greedy Sample Selection by choosing samples
    which maximize the importance score :math:`\pi`, which is the sum over
    the squares of the first :math:`k` components of the PCovR-modified
    left singular vectors.

    Parameters
    ----------

    mixing: float, default=0.5
            The PCovR mixing parameter, as described in PCovR as
            :math:`{\\alpha}`. Stored in :py:attr:`self.mixing`.

    recompute_every : int
                      number of steps after which to recompute the pi score
                      defaults to 1, if 0 no re-computation is done


    k : int
        number of eigenvectors to compute the importance score with, defaults to 1

    tolerance: float
         threshold below which scores will be considered 0, defaults to 1E-12

    n_to_select : int or float, default=None
        The number of selections to make. If `None`, half of the samples are
        selected. If integer, the parameter is the absolute number of selections
        to make. If float between 0 and 1, it is the fraction of the total dataset to
        select. Stored in :py:attr:`self.n_to_select`.

    score_threshold : float, default=None
        Threshold for the score. If `None` selection will continue until the
        n_to_select is chosen. Otherwise will stop when the score falls below the threshold.
        Stored in :py:attr:`self.score_threshold`.

    score_threshold_type : str, default="absolute"
        How to interpret the ``score_threshold``. When "absolute", the score used by
        the selector is compared to the threshold directly. When "relative", at each iteration,
        the score used by the selector is compared proportionally to the score of the first
        selection, i.e. the selector quits when ``current_score / first_score < threshold``.
        Stored in :py:attr:`self.score_threshold_type`.

    progress_bar: bool, default=False
              option to use `tqdm <https://tqdm.github.io/>`_
              progress bar to monitor selections. Stored in :py:attr:`self.report_progress`.

    full : bool, default=False
        In the case that all non-redundant selections are exhausted, choose
        randomly from the remaining samples. Stored in :py:attr:`self.full`.

    random_state: int or RandomState instance, default=0

    Attributes
    ----------

    X_current_ : ndarray (n_samples, n_features)
                  The original matrix orthogonalized by previous selections

    y_current_ : ndarray (n_samples, n_properties)
                The targets orthogonalized by a regression on
                the previous selections.

    n_selected_ : int
                  Counter tracking the number of selections that have been made
    X_selected_ : ndarray,
                  Matrix containing the selected samples, for use in fitting
    y_selected_ : ndarray,
                  In sample selection, the matrix containing the selected targets, for use in fitting

    """

    def __init__(
        self,
        mixing=0.5,
        recompute_every=1,
        k=1,
        tolerance=1e-12,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        super().__init__(
            selection_type="sample",
            mixing=mixing,
            recompute_every=recompute_every,
            k=k,
            tolerance=tolerance,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            score_threshold_type=score_threshold_type,
            progress_bar=progress_bar,
            full=full,
            random_state=random_state,
        )


def _directional_distance(equations, points):
    """
    Computes the distance of the points to the planes defined by the equations
    with respect to the direction of the first dimension.

    equations : ndarray of shape (n_facets, n_dim)
                each row contains the coefficienst for the plane equation of the form
                equations[i, 0]*x_1 + ...
                    + equations[i, -2]*x_{n_dim} = equations[i, -1]
                -equations[i, -1] is the offset
    points    : ndarray of shape (n_samples, n_dim)
                points to compute the directional distance from

    Returns
    -------
    directional_distance : ndarray of shape (nsamples, nequations)
                closest distance wrt. the first dimension of the point to the planes
                defined by the equations
    """
    orthogonal_distances = -(points @ equations[:, :-1].T) - equations[:, -1:].T
    return -orthogonal_distances / equations[:, :1].T


class DirectionalConvexHull:
    """
    Performs Sample Selection by constructing a Directional Convex Hull and determining the distance to the hull as outlined in the reference

    Parameters
    ----------

    low_dim_idx    : list of ints, default None
                   Indices of columns of X containing features to be used for the
                   directional convex hull construction (also known as the low-
                   dimensional (LD) hull). By default [0] is used.

    tolerance      : float, default=1.0E-12
                   Tolerance for the negative distances to the directional
                   convex hull to consider a point below the convex hull. Depending
                   if a point is below or above the convex hull the distance is
                   differently computed. A very low value can result in a completely
                   wrong distances. Distances cannot be distinguished from zero  up
                   to tolerance. It is recommended to leave the default setting.

    Attributes
    ----------

    high_dim_idx_   : list of ints
                    Indices of columns in data containing high-dimensional
                    features (i.e. those not used for the convex hull
                    construction)

    selected_idx_ : numpy.ndarray
                    Indices of datapoints that form the vertices of the
                    convex hull
    interpolator_high_dim_  : scipy.interpolate.interpnd.LinearNDInterpolator
                    Interpolater for the features in the high-
                    dimensional space

    References
    ----------
    .. [dch] A. Anelli, E. A. Engel, C. J. Pickard and M. Ceriotti,
             Physical Review Materials, 2018.
    """

    def __init__(self, low_dim_idx=None, tolerance=1e-12):
        self.low_dim_idx = low_dim_idx

        if low_dim_idx is None:
            self.low_dim_idx = [0]
        else:
            self.low_dim_idx = low_dim_idx
        self.tolerance = tolerance

    def fit(self, X, y):
        """
        Learn the samples that form the convex hull.

        Parameters
        ----------
        X        : ndarray of shape (n_samples, n_features)
                   Feature matrix of samples to use for constructing the convex
                   hull.
        y        : ndarray of shape (n_samples,)
                   Target values (property on which the convex hull should be
                   constructed, e.g. Gibbs free energy)

        Returns
        -------
        self : object
            Fitted scorer.
        """

        X, y = self._check_X_y(X, y)
        self.n_features_in_ = X.shape[1]

        if len(y.shape) == 1:
            y = y.reshape((len(y), 1))

        if (max(np.abs(self.low_dim_idx)) > X.shape[1]) and (
            min(self.low_dim_idx) >= 0
        ):
            raise ValueError(
                "One or more columns indexed with low_dim_idx is"
                " out of bounds with the dimensions of X."
            )

        self.high_dim_idx_ = np.setdiff1d(np.arange(X.shape[1]), self.low_dim_idx)

        # get number of dimensions for the convex (lower dimensional) hull construction
        n_low_dim_idx = len(self.low_dim_idx)

        # append features and target property to the same data matrix (for the
        # convex hull construction)
        convex_hull_data = np.zeros((X.shape[0], n_low_dim_idx + 1))
        convex_hull_data[:, :1] = y
        convex_hull_data[:, 1:] = X[:, self.low_dim_idx].copy()

        # create high-dimensional feature matrix
        high_dim_feats = X[:, self.high_dim_idx_]
        # get scipy convex hull
        self.convex_hull_ = ConvexHull(convex_hull_data, incremental=True)
        # get normal equations to the hull simplices
        y_normal = self.convex_hull_.equations[:, 0]

        # get vertices_idx of the convex hull
        directional_facets_idx = np.where(y_normal < 0)[0]
        self.directional_simplices_ = self.convex_hull_.simplices[
            directional_facets_idx
        ]
        self._directional_equations_ = self.convex_hull_.equations[
            directional_facets_idx
        ]

        self.selected_idx_ = np.unique(self.directional_simplices_.flatten())
        self._directional_points_ = convex_hull_data[self.selected_idx_]

        # required for the score_feature_matrix function
        self.interpolator_high_dim_ = _linear_interpolator(
            points=convex_hull_data[self.selected_idx_, 1:],
            values=high_dim_feats[self.selected_idx_],
        )

        return self

    @property
    def directional_vertices_(self):
        return self.selected_idx_

    def _check_X_y(self, X, y):
        return check_X_y(X, y, ensure_min_features=1, multi_output=False)

    def _check_is_fitted(self, X):
        check_is_fitted(
            self,
            [
                "high_dim_idx_",
                "interpolator_high_dim_",
                "selected_idx_",
            ],
        )
        n_features = X.shape[1]
        if n_features != self.n_features_in_:
            raise ValueError(
                f"X has {n_features} features, but {self.__class__.__name__} "
                f"is expecting {self.n_features_in_} features as input."
            )
        return True

    def score_samples(self, X, y):
        """
        Calculate the distance of the samples to the convex hull in the target
        direction y. Samples with a distance > 0 lie above the convex surface.
        Samples with a distance of zero lie on the convex surface. Samples with
        a distance value < 0 lie below the convex surface.

        Parameters
        ----------
        X    : ndarray of shape (n_samples, n_features)
               Feature matrix of samples to use for determining distance
               to the convex hull. Please note that samples provided should
               have the same dimensions (features) as used during fitting
               of the convex hull. The same column indices will be used for
               the low- and high-dimensional features.

        y    : ndarray of shape (n_samples,)
               Target values (property on which the convex hull should be
               constructed, e.g. Gibbs free energy)

        Returns
        -------
        dch_distance : ndarray of shape (n_samples, len(high_dim_idx_))
            The distance (residuals)  of samples to the convex hull in
            the higher-dimensional space.
        """
        X, y = self._check_X_y(X, y)
        self._check_is_fitted(X)

        # features used for the convex hull construction
        low_dim_feats = X[:, self.low_dim_idx]
        hull_space_data = np.hstack((y.reshape(-1, 1), low_dim_feats))

        # the X points projected on the convex surface
        return self._directional_convex_hull_distance(hull_space_data).reshape(y.shape)

    def _directional_convex_hull_distance(self, points):
        """
        Get the distance to the fitted directional convex hull in the target dimension y.

        If a point lies above the convex hull, it will have a positive distance.
        If a point lies below the convex hull, it will have a negative distance - this should only
        occur if you pass a point that the convex hull has not been fitted on in the `fit` function.
        If a point lies on the convex hull, it will have a distance of zero.

        Parameters
        ----------
        points : The points for which you would like to calculate the distance to the
                 fitted directional convex hull.
        """
        # directional distances to all plane equations
        all_directional_distances = _directional_distance(
            self._directional_equations_, points
        )
        # we get negative distances for each plane to check if any distance is below the threshold
        below_directional_convex_hull = np.any(
            all_directional_distances < -self.tolerance, axis=1
        )
        # directional distances to corresponding plane equation
        directional_distances = np.zeros(len(points))
        directional_distances[~below_directional_convex_hull] = np.min(
            all_directional_distances[~below_directional_convex_hull], axis=1
        )
        # some distances can be positive, so we take the max of all negative distances
        negative_directional_distances = all_directional_distances.copy()
        negative_directional_distances[all_directional_distances > 0] = -np.inf
        directional_distances[below_directional_convex_hull] = np.max(
            negative_directional_distances[below_directional_convex_hull], axis=1
        )
        return directional_distances

    def score_feature_matrix(self, X):
        """
        Calculate the distance (or more specifically, the residuals) of the
        samples to the convex hull in the high-dimensional space. Samples
        with a distance value of zero in all the higher dimensions lie on
        the convex hull.


        Parameters
        ----------
        X    : ndarray of shape (n_samples, n_features)
               Feature matrix of samples to use for determining distance
               to the convex hull. Please note that samples provided should
               have the same dimensions (features) as used during fitting
               of the convex hull. The same column indices will be used for
               the low- and high-dimensional features.

        Returns
        -------
        dch_distance : numpy.array of shape (n_samples, len(high_dim_idx_))
            The distance (residuals)  of samples to the convex hull in
            the higher-dimensional space.
        """
        X = check_array(X)
        self._check_is_fitted(X)

        # features used for the convex hull construction
        low_dim_feats = X[:, self.low_dim_idx]
        # HD features not used for the convex hull
        high_dim_feats = X[:, self.high_dim_idx_]

        if len(self.low_dim_idx) == 1:
            low_dim_feats = low_dim_feats.reshape(
                -1,
            )
        # interpolate features
        interpolated_high_dim_feats = self.interpolator_high_dim_(low_dim_feats)

        if np.any(np.isnan(interpolated_high_dim_feats)):
            warnings.warn(
                "There are samples in X with a low-dimensional part that is outside of the range of the convex surface. Distance will contain nans.",
                UserWarning,
            )

        # determine the distance between the original high-dimensional data and
        # interpolated high-dimensional data
        dch_distance = high_dim_feats - interpolated_high_dim_feats

        return dch_distance
